# Copyright (c) Meta Platforms, Inc. and affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import tempfile
import unittest

from hta.common.trace_file import (
    create_rank_to_trace_dict,
    read_trace,
    update_trace_rank,
    write_trace,
)
from hta.configs.config import logger


class TestTraceFile(unittest.TestCase):
    test_trace_data = {
        "distributedInfo": {"rank": 1},
        "traceEvents": [
            {
                "ph": "X",
                "cat": "cpu_op",
                "name": "aten::zeros",
                "pid": 2568503,
                "tid": 2568503,
                "ts": 1661938466265032,
                "dur": 25,
                "args": {
                    "Trace name": "PyTorch Profiler",
                    "Trace iteration": 0,
                    "External id": 1,
                    "Profiler Event Index": 0,
                },
            },
            {
                "ph": "X",
                "cat": "user_annotation",
                "name": "ProfilerStep#1009",
                "pid": 2568503,
                "tid": 2568503,
                "ts": 1661938466265087,
                "dur": 100298,
                "args": {
                    "Trace name": "PyTorch Profiler",
                    "Trace iteration": 0,
                    "External id": 5,
                    "Profiler Event Index": 4,
                },
            },
        ],
    }

    def setUp(self) -> None:
        self.trace_without_distributed_info = "tests/data/distributed_info_unavailable"
        self.trace_without_rank = "tests/data/rank_unavailable"
        self.trace_mixed_files = "tests/data/mixed_files"
        self.logger = logger

    def test_create_rank_to_trace_dict_without_distributed_info(self):
        with self.assertLogs(logger, level="ERROR") as cm:
            self.assertEqual(
                create_rank_to_trace_dict(self.trace_without_distributed_info),
                (True, {}),
            )
            self.assertIn("trace file does not have the rank", cm.output[0])

    def test_create_rank_to_trace_dict_without_rank(self) -> None:
        with self.assertLogs(logger, level="ERROR") as cm:
            self.assertEqual(
                create_rank_to_trace_dict(self.trace_without_rank), (True, {})
            )
            self.assertIn("trace file does not have the rank", cm.output[0])

    def test_create_rank_to_trace_dict_with_mixed_dir(self) -> None:
        with self.assertLogs(logger, level="ERROR") as cm:
            self.assertEqual(
                create_rank_to_trace_dict(self.trace_mixed_files),
                (True, {0: "tests/data/mixed_files/rank_non_gpu.json.gz"}),
            )
            self.assertIn("has the same rank", cm.output[0])

    def test_read_write_trace(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdirname:
            test_trace_file = os.path.join(tmpdirname, "test.json.gz")
            write_trace(TestTraceFile.test_trace_data, test_trace_file)
            read_trace_data = read_trace(test_trace_file)
            self.assertDictEqual(read_trace_data, TestTraceFile.test_trace_data)

    def test_update_trace_rank(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdirname:
            test_trace_file = os.path.join(tmpdirname, "test.json.gz")
            test_rank = 99
            write_trace(TestTraceFile.test_trace_data, test_trace_file)
            update_trace_rank(test_trace_file, test_rank)
            read_trace_data = read_trace(test_trace_file)
            self.assertTrue("distributedInfo" in read_trace_data)
            self.assertTrue("rank" in read_trace_data["distributedInfo"])
            self.assertEqual(read_trace_data["distributedInfo"]["rank"], test_rank)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

import os
import pathlib
import re
import sys

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.


sys.path.insert(0, os.path.abspath(".."))


def find_version(version_file_path) -> str:
    with open(version_file_path) as version_file:
        version_match = re.search(
            r"^__version_tuple__ = (.*)", version_file.read(), re.M
        )
        if version_match:
            ver_tup = eval(version_match.group(1))
            ver_str = ".".join([str(x) for x in ver_tup])
            return ver_str
        raise RuntimeError("Unable to find version tuple.")


# -- Project information -----------------------------------------------------

project = "Holistic Trace Analysis"
copyright = "2022, Meta Platforms Inc."
author = "Meta Platforms Inc."
repo_root_path = pathlib.Path(__file__).absolute().parents[1]
release = find_version(str(repo_root_path) + "/hta/version.py")

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.napoleon",  # support NumPy and Google style docstrings
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "sphinx.ext.githubpages",
    "sphinx.ext.doctest",
    "sphinx.ext.ifconfig",
]

source_suffix = {".rst": "restructuredtext"}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["source/_static"]



# Copyright (c) Meta Platforms, Inc. and affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import re
from typing import Tuple

import pandas as pd
import plotly.express as px

from hta.analyzers.timeline import _get_unique_values, plot_timeline_gpu_kernels
from hta.common.trace import Trace, TraceSymbolTable


def extract_iteration_info(trace: Trace) -> pd.DataFrame:
    """Extract the iteration information from a trace.

    Args:
          trace (Trace) : a trace object

    Returns:
        a DataFrame that contains all the iteration information with rank as a column.
    """
    ranks = [k for k in trace.get_all_traces().keys()]
    sym_map = trace.symbol_table.get_sym_id_map()
    sym_tab = trace.symbol_table.get_sym_table()

    def get_iter_nbr(profiler_step_name_id: int) -> int:
        iter_re = re.compile(r"ProfilerStep\s*#\s*(\d+)")
        m = iter_re.match(sym_tab[profiler_step_name_id])
        if m:
            return int(m.group(1))
        return -1

    def _extract_one_rank(df) -> pd.DataFrame:
        profiler_step_name_ids = [
            sym_map[k]
            for k in sorted([k for k in sym_map.keys() if k.startswith("ProfilerStep")])
        ]
        profiler_step_index = df["name"].isin(profiler_step_name_ids)
        p_steps = (
            df.loc[profiler_step_index, ["name", "ts", "dur"]]
            .copy()
            .rename(columns={"name": "profiler_step"})
        )
        p_steps["iter"] = p_steps["profiler_step"].apply(get_iter_nbr)
        p_steps["end"] = p_steps["ts"] + p_steps["dur"]
        p_steps.set_index("iter", inplace=True)
        p_steps.drop(["profiler_step"], axis=1, inplace=True)
        return p_steps

    df_iterations = pd.concat(
        [_extract_one_rank(trace.get_trace(r)) for r in ranks],
        keys=ranks,
        names=["rank", "iter"],
    ).reset_index()
    return df_iterations


def _compute_normalized_start_time_of_significant_comm_kernels(
    df: pd.DataFrame,
    symbol_table: TraceSymbolTable,
    visualize: bool = False,
    min_normalized_duration: float = 0.01,
) -> Tuple[pd.DataFrame, str]:
    """
    Computes the normalized start time of significant comm_kernels.

    Args:
        df (pd.DataFrame) : an input DataFrame with traces for selected ranks and iterations.
        symbol_table (TraceSymbolTable) : the Trace Symbol Table for encoding/decoding the symbols in the trace DataFrame
        visualize (bool): a flag to enable/disable visualizing the intermediate results.
        min_normalized_duration (float) : a ratio for filtering out operators/kernels which are unlikely
            to cause a rank to be a straggler.

    Return:
        Tuple[df: PdDataFrame, metric_name: str]
            df is a DataFrame which contains
            metric_name is a column name which can be used for straggler detection

    Note:
        + We break down straggler identification into two steps:
            1. compute metric for straggler detection
            2. detect straggler from the metric
        + This function provides a reference implementation for computing a metric for straggler detection. A user
            can choose a different metric and implement the corresponding metric computing function.
        + The metric used in this function is the starting time of the last most dominating communication kernel. It
            is computed in the following steps:
            1. find all communication kernels which pass the duration test, i.e.,
                kernel duration > mean_iter_time * min_normalized_duration (the default value is 1%)
            2. pick the last kernel on each stream
            3. select the stream-kernel combination which has the largest average standard deviation across ranks
        + This metric is defined for each rank and for each iteration (excluding the last iteration
            if it should be ignored).

        + One implicit assumption underlying this function is that candidate GPU communication kernels for straggler
            detection will involve a blocking all-to-all synchronization and thus end at about the same time.
            Therefore, the later a kernel start, the bigger overall delay it likely causes.

        + With more model-specific information, a user can create a metric easier to compute and understand.
    """
    # find all communication kernels
    sym_table = symbol_table.get_sym_table()
    sym_map = symbol_table.get_sym_id_map()
    comm_op_ids = [sym_map[s] for s in sym_table if s.startswith("ncclKernel")]
    iterations = _get_unique_values(df, "iteration")
    ranks = _get_unique_values(df, "rank")
    df = df.loc[df["iteration"].isin(iterations)]

    # filter out short communication kernels with a duration threshold
    n_iters = len(iterations)
    t_min = df["ts"].min()
    mean_iter_time = ((df["ts"] + df["dur"]).max() - t_min) / float(n_iters)
    min_duration = mean_iter_time * min_normalized_duration
    long_comm_kernels = df.loc[
        (df["stream"] > 0)
        & (df["iteration"] > 0)
        & (df["dur"] >= min_duration)
        & (df["name"].isin(comm_op_ids))
    ]
    if visualize:  # pragma: no cover
        plot_timeline_gpu_kernels(
            f"Timeline of Communication Kernels Longer Than {min_normalized_duration * 100:.2f}% of Iteration Time\n",
            long_comm_kernels,
            symbol_table,
            ranks=ranks,
            duration_threshold=2000,
        )

    # select the last long communication kernel for all combinations of (rank, stream, iteration, kernel_name)
    last_long_comm_kernels = long_comm_kernels.groupby(
        ["rank", "stream", "iteration", "name"], as_index=False
    ).last()

    # select the (stream, name) combination whose duration has the largest mean standard deviation across all ranks
    metric_name: str = "normalized_start_time"
    last_long_comm_kernels[metric_name] = (
        last_long_comm_kernels["ts"] - t_min
    ) / mean_iter_time
    last_long_comm_kernels["duration"] = last_long_comm_kernels["dur"] / mean_iter_time
    duration_diff_across_ranks = last_long_comm_kernels.groupby(
        ["stream", "iteration", "name"]
    )["duration"].std()
    average_duration_diff = duration_diff_across_ranks.groupby(
        ["stream", "name"]
    ).mean()
    (stream, name) = average_duration_diff.idxmax()

    # Filter last_longer_comm_kernels
    candidate_metric_kernels = last_long_comm_kernels.loc[
        (last_long_comm_kernels["stream"].eq(stream))
        & (last_long_comm_kernels["name"].eq(name))
    ]

    if visualize:  # pragma: no cover
        plot_timeline_gpu_kernels(
            f"Timeline of Candidate Kernels (Iterations={iterations})",
            candidate_metric_kernels,
            symbol_table,
            ranks=ranks,
        )

    return candidate_metric_kernels, metric_name


def _get_top_k_stragglers_with_metric(
    metric_df: pd.DataFrame,
    metric_name: str,
    n_stragglers: int = 2,
    visualize: bool = False,
) -> pd.Series:
    """
    Get the top k potential straggler ranks.

    Args:
          metric_df (pd.DataFrame) : an input DataFrame which contains the performance metrics
            for determining whether a particular rank could be a straggler.
          metric_name (str) : a column name to select the metric from the DataFrame.
          n_stragglers (int) : how many candidate stragglers to use in this algorithm?
          visualize (bool) : a flag to enable/disable visualizing the results.

    Returns:
        A Series of integers with ranks as the index and the values indicating whether a rank is a potential straggler.
        If the value for a rank <r> is positive, then rank <r> is a potential straggler;
        otherwise it is treated as a non-straggler.

    Notes:
        + This function is built upon the metric derived for straggler detection and applies a candidate selection
            algorithm to select potential straggler candidates.
        + The algorithm implemented here is a k late start candidate which choose the top k candidates ranked by the
            metric values.
        + This function consider two scenarios:
            (1) metric_df only contains metric for one iteration;
            (2) metric_df contains metric for multiple iterations.
    """
    if n_stragglers <= 0:
        n_stragglers = 1
    n_iterations = (
        metric_df["iteration"].nunique() if "iteration" in metric_df.columns else 1
    )
    with_iteration = True if n_iterations > 1 else False
    if not with_iteration:
        metric = metric_df[["rank", metric_name]].copy()
        threshold = metric.sort_values(by=metric_name, ascending=False)[
            metric_name
        ].tolist()[n_stragglers - 1]
        metric["is_straggler"] = metric[metric_name].apply(
            lambda x: 1 if x >= threshold else 0
        )
        data_columns = ["rank", metric_name]
        results = (
            metric[["rank", "is_straggler"]].copy().set_index("rank")["is_straggler"]
        )
    else:
        metric = metric_df[["rank", "iteration", metric_name]].copy()
        metric["is_straggler"] = 0
        for iteration, group in metric.groupby("iteration"):
            if iteration < 0:
                continue
            threshold = group.sort_values(by=metric_name, ascending=False)[
                metric_name
            ].tolist()[n_stragglers - 1]
            metric.loc[group.index, "is_straggler"] = group[metric_name].apply(
                lambda x: 1 if x >= threshold else 0
            )
        data_columns = ["rank", "iteration", metric_name]
        results = metric.groupby("rank")["is_straggler"].sum()

    metric["is_straggler"] = metric["is_straggler"].apply(
        lambda x: "Yes" if x > 0 else "No"
    )
    for col in data_columns:
        metric[col] = pd.to_numeric(metric[col])

    color_map = {"Yes": "red", "No": "blue"}
    if visualize:  # pragma: no cover
        fig = px.bar(
            metric,
            x="rank",
            y=metric_name,
            facet_col="iteration" if with_iteration else None,
            color="is_straggler",
            color_discrete_map=color_map,
            hover_data=data_columns,
            width=300 * n_iterations,
            title="Potential Stragglers",
        )
        fig.show()

    return results


def find_stragglers_with_late_start_comm_kernels(
    df: pd.DataFrame,
    symbol_table: TraceSymbolTable,
    n_stragglers: int = 2,
    visualize: bool = False,
) -> pd.Series:
    """Identify potential stragglers from the traces.

    Args:
          df (pd.DataFrame) : an input DataFrame with traces for selected ranks and iterations.
          symbol_table (TraceSymbolTable) : the Trace Symbol Table for encoding/decoding the symbols in the trace DataFrame
          n_stragglers (int) : how many candidate stragglers to use in this algorithm?
          visualize (bool) : a flag to enable/disable visualizing the results.

    Returns:
        stragglers (pd.Series): a Series of which ranks are the index and values are the number of times
            that a rank is detected as a potential straggler. For example, given a rank r,
            if stragglers[r] > 0, then rank <r> is detected as a straggler for stragglers[r] times.
    """

    metric_df, metric_name = _compute_normalized_start_time_of_significant_comm_kernels(
        df, symbol_table, visualize
    )
    stragglers = _get_top_k_stragglers_with_metric(
        metric_df, metric_name, visualize=visualize, n_stragglers=n_stragglers
    )
    return stragglers



# Copyright (c) Meta Platforms, Inc. and affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import multiprocessing as mp
import unittest
from typing import List, Set

from hta.common.trace import TraceSymbolTable


class SymbolDecoder:
    def __init__(self, symbol_table: TraceSymbolTable):
        self.sym_table = symbol_table.get_sym_table()
        self.sym_id_map = symbol_table.get_sym_id_map()

    def __call__(self, idx: int) -> str:
        return self.sym_table[idx]


def check_symbol_table(st: TraceSymbolTable) -> bool:
    sym_id_map = st.get_sym_id_map()
    sym_table = st.get_sym_table()
    two_way_map_consistency = [sym_table[i] == s for s, i in sym_id_map.items()]
    return all(two_way_map_consistency)


class TraceSymbolTableTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.symbols_1: List[str] = ["a", "b", "c", "b1"]
        self.symbols_2: List[str] = ["a", "b", "c", "b2", "d1"]
        self.symbols_3: List[str] = ["a", "b", "f", "b3", "d2"]
        ss: Set[str] = set()
        self.symbols_list = [self.symbols_1, self.symbols_2, self.symbols_3]
        for s in self.symbols_list:
            ss = ss.union(set(s))
        self.symbols: List[str] = sorted(list(ss))

    def test_add_symbols_single_process(self):
        st = TraceSymbolTable()
        for symbols in [self.symbols_1, self.symbols_2, self.symbols_3]:
            st.add_symbols(symbols)

        self.assertListEqual(sorted(st.get_sym_table()), self.symbols)
        self.assertTrue(check_symbol_table(st))

    def test_add_symbols_multi_processing(self):
        st = TraceSymbolTable()
        st.add_symbols_mp(self.symbols_list)

        self.assertListEqual(sorted(st.get_sym_table()), self.symbols)
        self.assertTrue(check_symbol_table(st))

    def test_query_symbols_multi_processes(self):
        st = TraceSymbolTable()
        st.add_symbols_mp(self.symbols_list)
        decoder = SymbolDecoder(st)
        indices = [i for i, _ in enumerate(st.get_sym_table())]

        np = 4
        with mp.Pool(np) as pool:
            decoded_symbols = pool.map(decoder, indices)
        pool.join()
        pool.close()
        self.assertEqual(len(decoded_symbols), len(self.symbols))

        sym_id_map = st.get_sym_id_map()
        is_consistent = [
            sym_id_map[sym] == idx for (sym, idx) in zip(decoded_symbols, indices)
        ]
        self.assertTrue(all(is_consistent))


if __name__ == "__main__":  # pragma: no cover
    unittest.main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import json
import unittest

from hta.configs.config import HtaConfig
from hta.configs.default_values import DEFAULT_CONFIG_FILENAME


class HtaConfigTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.test_config_path = "/tmp/test_config.json"
        self.test_config = {
            "a": 1,
            "b": ["s", "t"],
            "c": {"c1": 2, "c2": {"c21": 10.0}},
        }
        with open(self.test_config_path, "w+") as fp:
            json.dump(self.test_config, fp)

    def test_get_default_paths(self):
        paths = HtaConfig.get_default_paths()
        self.assertEqual(
            len(paths), 3, f"expect the default file paths to be 3 but got {len(paths)}"
        )
        self.assertTrue(
            all([str(path).endswith(DEFAULT_CONFIG_FILENAME) for path in paths])
        )

    def test_constructor_no_config_file(self):
        config = HtaConfig(load_default_paths=False)
        self.assertDictEqual(config.get_config(), {})

    def test_constructor_one_config_file(self):
        config = HtaConfig(self.test_config_path, load_default_paths=False)
        self.assertEqual(config.get_config(), self.test_config)

    def test_get_config_file_paths(self):
        config = HtaConfig(self.test_config_path, load_default_paths=False)
        paths = config.get_config_file_paths()
        self.assertListEqual(paths, [self.test_config_path])

    def test_get_config_all(self):
        config = HtaConfig(self.test_config_path, load_default_paths=False)
        config_values = config.get_config()
        self.assertDictEqual(config_values, self.test_config)

    def test_get_config_one_level(self):
        config = HtaConfig(self.test_config_path, load_default_paths=False)
        self.assertEqual(config.get_config("a"), self.test_config["a"])
        self.assertListEqual(config.get_config("b"), self.test_config["b"])
        self.assertDictEqual(config.get_config("c"), self.test_config["c"])

    def test_get_config_multiple_levels(self):
        config = HtaConfig(self.test_config_path, load_default_paths=False)
        self.assertDictEqual(config.get_config("c"), self.test_config["c"])
        self.assertEqual(config.get_config("c.c1"), self.test_config["c"]["c1"])
        self.assertEqual(
            config.get_config("c.c2.c21"), self.test_config["c"]["c2"]["c21"]
        )
        self.assertIsNone(config.get_config("d"))
        self.assertIsNone(config.get_config("c.c2.c22"))
        self.assertIsNone(config.get_config("c.c1.c3"))

    def test_get_config_default_values(self):
        config = HtaConfig(self.test_config_path, load_default_paths=False)
        self.assertEqual(config.get_config("c", 10), self.test_config["c"])
        self.assertEqual(config.get_config("d", 10), 10)


if __name__ == "__main__":  # pragma: no cover
    unittest.main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from collections import defaultdict
from typing import Dict, List, TYPE_CHECKING

import pandas as pd
import plotly.express as px

from hta.utils.utils import get_kernel_type, KernelType, merge_kernel_intervals

# import statement used without the "if TYPE_CHECKING" guard will cause a circular
# dependency with trace_analysis.py causing mypy to fail and should not be removed.
if TYPE_CHECKING:
    from hta.trace import Trace


class CommunicationAnalysis:
    def __init__(self):
        pass

    @classmethod
    def get_comm_comp_overlap(cls, t: "Trace", visualize: bool = True) -> pd.DataFrame:
        """
        Communication analysis implementation. See `get_comm_comp_overlap` in `trace_analysis.py` for details.
        """
        sym_table = t.symbol_table.get_sym_table()

        def get_comm_comp_overlap_value(trace_df: pd.DataFrame) -> float:
            """
            Compute the overlap percentage between communication and computation kernels for one rank.
            """
            gpu_kernels = trace_df[trace_df["stream"].ne(-1)].copy()
            gpu_kernels["kernel_type"] = gpu_kernels[["name"]].apply(
                lambda x: get_kernel_type(sym_table[x["name"]]), axis=1
            )

            # Isolate communication and computation kernels and merge each one of them.
            comp_kernels = merge_kernel_intervals(
                gpu_kernels[
                    gpu_kernels["kernel_type"].eq(KernelType.COMPUTATION.name)
                ].copy()
            )
            comm_kernels = merge_kernel_intervals(
                gpu_kernels[
                    gpu_kernels["kernel_type"].eq(KernelType.COMMUNICATION.name)
                ].copy()
            )

            # When a communication kernel starts and ends, the cumulative status is changed by 1 and -1;
            # when a computation kernel starts and ends, the cumulative status is changed by 2 and -2.
            status_df = (
                pd.concat(
                    [
                        comm_kernels.melt(var_name="status", value_name="time").replace(
                            {"ts": 1, "end": -1}
                        ),
                        comp_kernels.melt(var_name="status", value_name="time").replace(
                            {"ts": 2, "end": -2}
                        ),
                    ]
                )
                .sort_values(by="time")
                .reset_index(drop=True)
            )
            status_df["running"] = status_df["status"].cumsum()
            # Time intervals when status is 3 indicate overlapping communication and computation kernels.
            overlap = status_df[status_df["running"].eq(3)]
            shifted_overlap = overlap.merge(
                status_df.shift(-1).dropna(), left_index=True, right_index=True
            )
            return (shifted_overlap["time_y"] - shifted_overlap["time_x"]).sum() / (
                comm_kernels["end"] - comm_kernels["ts"]
            ).sum()

        result: Dict[str, List[float]] = defaultdict(list)
        for rank, trace_df in t.traces.items():
            result["rank"].append(rank)
            result["comp_comm_overlap_ratio"].append(
                get_comm_comp_overlap_value(trace_df)
            )
        result_df = pd.DataFrame(result)
        result_df["comp_comm_overlap_pctg"] = round(
            100 * result_df["comp_comm_overlap_ratio"], 2
        )

        if visualize:  # pragma: no cover
            fig = px.bar(
                result_df,
                x="rank",
                y="comp_comm_overlap_ratio",
                title="Computation-Communication Overlap",
                labels={
                    "rank": "Rank",
                    "comp_comm_overlap_ratio": "Computation-Communication Overlap Percentage",
                },
            )

            fig.update_layout(yaxis_tickformat=".2%")
            fig.show()

        return result_df[["rank", "comp_comm_overlap_pctg"]]



# Copyright (c) Meta Platforms, Inc. and affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from collections import defaultdict
from enum import auto, Flag
from typing import Any, Callable, Dict, List, Optional, Tuple

import pandas as pd

from hta.analyzers.breakdown_analysis import BreakdownAnalysis
from hta.analyzers.communication_analysis import CommunicationAnalysis
from hta.analyzers.cuda_kernel_analysis import CudaKernelAnalysis
from hta.analyzers.cupti_counter_analysis import CuptiCounterAnalysis
from hta.analyzers.straggler import find_stragglers_with_late_start_comm_kernels
from hta.analyzers.straggler_analysis import StragglerAnalysis
from hta.analyzers.trace_counters import TraceCounters
from hta.common.trace import Trace
from hta.configs.config import logger
from hta.configs.default_values import DEFAULT_TRACE_DIR


class TimeSeriesTypes(Flag):
    QUEUE_LENGTH = auto()
    MEMCPY_BANDWIDTH = auto()


class TraceAnalysis:
    def __init__(
        self,
        trace_files: Optional[Dict[int, str]] = None,
        trace_dir: str = DEFAULT_TRACE_DIR,
    ):
        self.t = Trace(trace_files, trace_dir)
        self.t.load_traces()
        assert self.t.is_parsed is True

    def get_comm_comp_overlap(self, visualize: bool = True) -> pd.DataFrame:
        r"""
        Compute the communication-computation overlap percentage for each rank.

        Args:
            visualize (bool): Set to True to display the graph. Default = True.

        Returns:
            pd.DataFrame
                A dataframe containing the communication computation overlap percentage for each rank.
        """
        return CommunicationAnalysis.get_comm_comp_overlap(self.t, visualize)

    def get_profiler_steps(self) -> List[int]:
        r"""
        Get the list of profiler steps.

        Args:
            None

        Returns:
            List[int]
                A list of profiler steps.
        """
        return StragglerAnalysis.get_profiler_steps(self.t)

    def get_potential_stragglers(
        self,
        profiler_steps: Optional[List[int]] = None,
        num_candidates: int = 2,
        visualize: bool = False,
        straggler_identification_impl: Callable[
            ..., pd.Series
        ] = find_stragglers_with_late_start_comm_kernels,
    ) -> List[int]:
        r"""
        Identify potential stragglers based on a pre-defined metric computed from the trace.

        Args:
            profiler_steps (List[int]) : A list of profiler steps used in the straggler analysis. If None, then use all
                                         iterations.
            num_candidates (int): Number of straggler candidates straggler analysis should look for
            visualize (bool): Set to True to visualize intermediate results
            straggler_identification_impl: A function or method which implements a straggler identification algorithm.
                                           The function takes a DataFrame, a TraceSymbolTable, num_candidates, and
                                           visualize as input and returns a Series with rank as the index and the
                                           value as an indicator whether a rank is a potential straggler.

        Returns:
            List[int]
                The list of ranks which can be potential stragglers.

        Notes:
            1. Find all communication kernels which pass the duration test, i.e., kernel duration > mean_iter_time * min_normalized_duration (the default value is 1%).
            2. Pick the last kernel for each combination of (iteration, rank, stream, kernel).
            3. For each combination of (iteration, stream, kernel), compute the standard deviation of the last kernel duration w.r.t. ranks.
            4. Compute the mean value of the last kernel duration standard deviation w.r.t. ranks across all available iterations. If there is only one iteration, the mean value would just be the std computed at 3.
            5. Select the (stream, kernel) combination which has the largest mean value computed at 4.
            6. Select the normalized start time for the (stream, kernel) combination as the candidate metric for detecting whether a rank will be a straggler.

            + This method forwards the analysis to the function hta.straggler.default_straggler_identification.
            + Because most distributed training jobs have one or multiple blocking synchronizations within an iteration, we define a straggler as the trainer which starts later than others and thus causes the largest slowdown of the overall training process.
            + Because empirical knowledge shows that all-to-all collective communication operators are most likely to have the largest impact on the training performance, this straggler analysis computes a default straggler metric as follows:
            + This metric is defined for each rank and for each iteration (excluding the last iteration if it should be ignored).
            + The metrics is passed into a straggler detection algorithm which picks the top k ranks ordered by the metric computed above for each iteration, where k is num_candidates.
            + For multiple iterations, it is possible that different iteration may return a different set of potential stragglers. The identify_potential_stragglers() will return the counts of how many times each rank is found to be a potential straggler.
        """
        return StragglerAnalysis.get_potential_stragglers(
            self.t,
            profiler_steps,
            num_candidates,
            visualize,
            straggler_identification_impl,
        )

    def get_gpu_kernel_breakdown(
        self,
        visualize: bool = True,
        duration_ratio: float = 0.8,
        num_kernels: int = 10,
        include_memory_kernels: bool = True,
        image_renderer: str = "notebook",
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        r"""
        Summarizes the time spent by each kernel and by kernel type. Outputs the following graphs:

        1. Pie chart indicating the percentage of time taken by each kernel type.
        2. Pie charts showing the most time consuming kernels for each rank for each kernel type.
        3. Bar graphs showing the average duration for the most time consuming kernels for each rank and each kernel type.

        Args:
            visualize (boolean): Set to True to display the graphs. Default = True.
            duration_ratio (float): Floating point value between 0 and 1 specifying the ratio of time taken
                                    by top COMM/COMP/MEMORY kernels. Default = 0.8.
            num_kernels (int): Maximum number of COMM/COMP/MEMORY kernels to show. Default = 10.
            include_memory_kernels (bool): Whether to include MEMORY kernels in the analysis. Default = True.
            image_renderer (str): Set to ``notebook`` when using jupyter and ``jupyterlab`` when using jupyter-lab.
                To see all available options execute: ``import plotly; plotly.io.renderers`` in a python shell.

        Returns:
            Tuple[pd.DataFrame, pd.DataFrame]
                Returns two dataframes. The first dataframe shows the percentage of time spent by kernel type.
                The second dataframe shows the min, max, mean, standard deviation, total time taken by each
                kernel on each rank. This dataframe will be summarized based on values of ``duration_ratio``
                and ``num_kernels``. If both ``duration_ratio`` and ``num_kernels`` are specified,
                ``num_kernels`` takes precedence.
        """

        return BreakdownAnalysis.get_gpu_kernel_breakdown(
            self.t,
            visualize,
            duration_ratio,
            num_kernels,
            include_memory_kernels,
            image_renderer,
        )

    def get_temporal_breakdown(self, visualize: bool = True) -> pd.DataFrame:
        r"""
        Compute the idle time, compute time and non-compute time for each rank. Time is measured in
        nanoseconds (ns). non-compute time is defined as the total time the GPU is not executing a
        compute operation such as data transfers, copying to/from memory and communication collectives.
        (In the strictest sense communication collectives do some compute but we classify it as a
        communication operation).

        Args:
            visualize (bool): Set to True to display the graphs. Default = True.

        Returns:
            pd.DataFrame
                A dataframe containing the raw value and percentage of idle time, compute time and non-compute
                time for each rank.
        """
        return BreakdownAnalysis.get_temporal_breakdown(self.t, visualize)

    def get_frequent_cuda_kernel_sequences(
        self,
        operator_name: str,
        output_dir: str,
        min_pattern_len: int = 3,
        rank: int = 0,
        top_k: int = 5,
        visualize: bool = False,
        compress_other_kernels: bool = True,
    ) -> pd.DataFrame:
        r"""
        Computes the most frequent CUDA kernel sequences originating from the CPU op with name
        ``operator_name``. Generates a dataframe summarizing the sequence of kernels, their frequency
        and total time taken. Additionally, writes a new trace file to ``output_dir`` with the
        top_k frequent patterns overlaid on top of the original trace file.

        Args:
            operator_name (str): Name of the operator from which the CUDA kernels are launched.
            output_dir (str): Output folder path containing the new trace file with overlaid top k
                              frequent patterns.
            min_pattern_len (int): Minimum length of the CUDA kernel sequences that should be identified. Default = 3.
            rank (int): Rank on which the analysis is performed. Default = 0.
            top_k (int): top_k patterns in terms of frequency to be visualized and overlaid. Default = 5.
            visualize (bool): Whether to show the histogram of top_k frequent patterns inline. Default = True.
            compress_other_kernels (bool): Should the names and args for other kernels not belonging to any
                                           frequent patterns be compressed to save memory in the overlaid
                                           trace file. Default = True.

        Returns:
            pd.DataFrame
                A dataframe with frequent cuda kernel sequences and their frequencies.
        """
        return CudaKernelAnalysis.get_frequent_cuda_kernel_sequences(
            self.t,
            operator_name,
            output_dir,
            min_pattern_len,
            rank,
            top_k,
            visualize,
            compress_other_kernels,
        )

    def get_cuda_kernel_launch_stats(
        self,
        ranks: Optional[List[int]] = None,
        runtime_cutoff: int = 50,
        launch_delay_cutoff: int = 100,
        include_memory_events: bool = True,
        visualize: bool = True,
    ) -> Dict[int, pd.DataFrame]:
        r"""
        For each event launched on the GPU there is a corresponding scheduling event on the CPU.
        These events are linked by a common correlation id. This feature calculates the duration
        of the CPU op, GPU kernel and the launch delay (difference between gpu kernel starting
        and cpu op ending for each correlation id on the specified rank(s). This function finds:
            1. GPU events with a shorter duration than the corresponding CPU events.
            2. CPU runtime events with a large duration i.e. outliers. The outliers are defined using
               the ``runtime_cutoff`` value.
            3. CPU events which have a large launch delay i.e. launch delay outliers. The launch delay
               outliers are defined using the ``launch_delay_cutoff`` value.

        Args:
            ranks (List[int]): List of ranks on which to run the analysis. Default = [0].
            runtime_cutoff (int): Duration in microseconds to determine outliers for cuda runtime events.
                Default = 50 microseconds.
            launch_delay_cutoff (int): Duration in microseconds to determine outliers for launch delay.
                Default value is 100 microseconds.
            include_memory_events (bool): Toggle to include cudaMemcpyAsync and cudaMemsetAsync events.
                Default = True.
            visualize (bool): Toggle to display the generated graphs. Default = True.

        Returns:
            Dict[int, pd.DataFrame]
                The function returns a dictionary of dataframes. The key corresponds to the rank and the value is
                a dataframe containing the cpu_duration, gpu_duration and launch_delay for each correlation id.
        """
        if ranks is None:
            ranks = [0]

        return CudaKernelAnalysis.cuda_kernel_launch_stats(
            self.t,
            ranks,
            runtime_cutoff,
            launch_delay_cutoff,
            include_memory_events,
            visualize,
        )

    def generate_trace_with_counters(
        self,
        time_series: Optional[TimeSeriesTypes] = None,
        ranks: Optional[List[int]] = None,
        output_suffix: str = "_with_counters",
    ) -> None:
        r"""
        Adds a set of time series to the trace in order to aid debugging traces. Creates a new trace file
        for each requested rank with the a suffix '_with_counters.json'. The following time series are
        available in TimeSeriesTypes flag type.

        1. Queue length - adds a time series to the trace indicating the size of the queue at any given time on each CUDA stream.
        2. Memory copy bandwidth - adds a time series to the trace indicating the memory bandwidth used for device to host, host to device and device to device operations.

        Either or both of the above can be enabled.

        Args:
            time_series (Flag): Used to set the requested time series. Available values are
                TimeSeriesTypes.QUEUE_LENGTH and TimeSeriesTypes.MEMCPY_BANDWIDTH. By default
                both time series are added to the trace.
            ranks (List[int]): List of ranks to generate the counters for. Default = [0].
            output_suffix (str): Suffix to add to the trace file. Default = '_with_counters.json.gz'

        Returns:
            None
        """
        if ranks is None or len(ranks) == 0:
            ranks = [0]

        if time_series is None:
            time_series = (
                TimeSeriesTypes.QUEUE_LENGTH | TimeSeriesTypes.MEMCPY_BANDWIDTH
            )

        counter_events: Dict[int, List[Dict[str, Any]]] = defaultdict(list)
        if output_suffix == "":
            output_suffix = "_with_counters"

        def add_time_series(
            series_dict: Dict[int, pd.DataFrame], counter_name: str, counter_col: str
        ):
            nonlocal counter_events
            """accept a rank -> time series dict and append it"""
            for rank, series in series_dict.items():
                if "stream" in series.columns:
                    series.rename(columns={"stream": "id"}, inplace=True)
                ce = self.t.convert_time_series_to_events(
                    series, counter_name, counter_col
                )
                counter_events[rank].extend(ce)

        if TimeSeriesTypes.QUEUE_LENGTH in time_series:
            add_time_series(
                series_dict=TraceCounters.get_queue_length_time_series(self.t, ranks),
                counter_name="Queue Length",
                counter_col="queue_length",
            )
        if TimeSeriesTypes.MEMCPY_BANDWIDTH in time_series:
            add_time_series(
                series_dict=TraceCounters.get_memory_bw_time_series(self.t, ranks),
                counter_name="Memcpy BW",
                counter_col="memory_bw_gbps",
            )

        for rank, ev_list in counter_events.items():
            raw_trace_content = self.t.get_raw_trace_for_one_rank(rank=rank)
            raw_trace_content["traceEvents"].extend(ev_list)
            output_file = self.t.trace_files[rank].replace(
                ".json", f"{output_suffix}.json"
            )
            logger.info(f"Writing trace with counters for rank {rank} to {output_file}")
            self.t.write_raw_trace(output_file, raw_trace_content)

    def get_queue_length_summary(
        self,
        ranks: Optional[List[int]] = None,
    ) -> Optional[pd.DataFrame]:
        r"""
        Queue length is defined as the number of outstanding CUDA operations on a stream. This
        functions calculates the summary statistics for the queue length on each CUDA stream for
        the specified ranks.

        Args:
            ranks (List[int]): List of ranks for which to queue length summary is calculated. Default = [0].

        Returns:
            pd.DataFrame or None
                A dataframe summarizing the queue length statistics. The dataframe contains count,
                min, max, standard deviation, 25th, 50th and 75th percentiles.
                The function returns None when the dataframe is empty.
        """
        return TraceCounters.get_queue_length_summary(self.t, ranks)

    def get_queue_length_time_series(
        self,
        ranks: Optional[List[int]] = None,
    ) -> Dict[int, pd.DataFrame]:
        r"""
        Queue length is defined as the number of outstanding CUDA operations on a stream. This
        function calculates the time series for the queue length on each CUDA stream for the
        specified ranks.

        Args:
            ranks (List[int]): List of ranks for which the queue length time series is generated. Default = [0].

        Returns:
            Dict[int, pd.DataFrame]
                Returns a dictionary whose key is the rank and value is a dataframe of queue length
                counter events. The following fields are in each row of the dataframe: ts (timestamp), pid (process id),
                tid (thread id), stream, and queue length.
        """
        return TraceCounters.get_queue_length_time_series(self.t, ranks)

    def get_memory_bw_summary(
        self,
        ranks: Optional[List[int]] = None,
    ) -> pd.DataFrame:
        r"""
        Summarizes the memory bandwidth statistics for memory copy and memset operations. This includes memory
        bandwidth for copies from Device to Host, Host to Device and Device to Device transfers. Note, this does
        not include memory bandwidth used by compute/communication kernels.

        Args:
            ranks (List[int]): List of ranks for which memory bandwidth is calculated. Default = [0].

        Returns:
            pd.DataFrame or None
                A dataframe containing the summary statistics. The dataframe includes count, min, max, standard deviation,
                25th, 50th and 75th percentiles of memory copy/memset operations.
                The function returns None when the dataframe is empty.
        """
        return TraceCounters.get_memory_bw_summary(self.t, ranks)

    def get_memory_bw_time_series(
        self,
        ranks: Optional[List[int]] = None,
    ) -> Dict[int, pd.DataFrame]:
        r"""
        Calculates the time series for memory copy bandwidth used by memcpy and memset operations in GB/s. The
        memory bandwidth is calculated for host to device, device to host and device to device copies. Note, this
        does not include memory bandwidth used by computation or communication kernels.

        Args:
            ranks (List[int]): List of ranks for which the memory bandwidth time series is generated. Default = [0].

        Returns:
            Dict[int, pd.DataFrame]
                Returns a dictionary whose key is the rank and value is a dataframe of memory bandwidth
                counter events. The following fields are in each row of the dataframe: ts (timestamp), pid (process id),
                tid (thread id), name (memcpy/memset), and memory bandwidth in GB/s.
        """
        return TraceCounters.get_memory_bw_time_series(self.t, ranks)

    def get_idle_time_breakdown(
        self,
        ranks: Optional[List[int]] = None,
        streams: Optional[List[int]] = None,
        visualize: bool = True,
        visualize_pctg: bool = True,
        show_idle_interval_stats=False,
        consecutive_kernel_delay: int = 30,
    ) -> Tuple[pd.DataFrame, Optional[pd.DataFrame]]:
        r"""
        GPU is considered idle when no kernel is running on it. Idle time is broken down into 3 categories.

        1. Host wait time: a GPU or stream is idle because the CPU thread has not enqueued enough kernels to
           keep it occupied.
        2. Kernel wait time: This is the duration between kernels and is considered an overhead of
           launching multiple small kernels. We use the following heuristic to classify the duration as kernel wait:
           duration between consecutive kernels < ``consecutive_kernel_delay``.
        3. Other wait time: In this case the idle time is attributed to an unknown cause. For example, a compute
           kernel could be waiting for a CUDA event from a communication kernel to complete.

        Args:
            ranks (List[int]): List of ranks for which idle time breakdown is computed. Default = [0].
            streams (List[int]): List of streams to provide analysis for. Defaults to all streams.
            visualize (bool): Set to True to show the graph. Default = True.
            visualize_pctg (bool): Show relative percentage across streams. Default = True.
            show_idle_interval_stats (bool): Returns statistics of the idle intervals like the min, max
               and median of idle intervals between kernels on a CUDA stream, also broken down by
               the idleness category. Default = False.
            consecutive_kernel_delay (int): Configures the threshold under which we consider gaps between
                kernels to be due to realistic delays in launching back to back kernels on the GPU.
                Default = 30 nanoseconds.

        Returns:
            Tuple[pd.DataFrame, Optional[pd.DataFrame]]
                A tuple of dataframes. The first dataframe contains the idle time category and duration for each stream
                on each rank. The second dataframe contains the summary statistics (count, min, max, mean, standard deviation,
                25th, 50th, 75th percentile) for each idle category for each stream on each rank.
        """
        if ranks is None or len(ranks) == 0:
            ranks = [0]

        idle_time_df_list: List[pd.DataFrame] = []
        interval_df_list: List[pd.DataFrame] = []
        for rank in ranks:
            idle_time_r_df, interval_r_df = BreakdownAnalysis.get_idle_time_breakdown(
                self.t,
                consecutive_kernel_delay,
                rank,
                streams,
                visualize,
                visualize_pctg,
                show_idle_interval_stats,
            )
            idle_time_df_list.append(idle_time_r_df)
            if interval_r_df is not None:
                interval_df_list.append(interval_r_df)

        return (
            pd.concat(idle_time_df_list),
            pd.concat(interval_df_list) if show_idle_interval_stats else None,
        )

    def get_cupti_counter_data_with_operators(
        self,
        ranks: Optional[List[int]] = None,
    ) -> List[pd.DataFrame]:
        r"""Performance counters provide insights on how to speed up GPU
        kernels. The PyTorch Profiler has a lightweight API [CUPTI Range
        Profiler API](https://docs.nvidia.com/cupti/r_main.html#r_profiler)
        that enables users to monitor performance counters from the device.

        When the CUPTI Profiler mode is enabled then PyTorch will emit the
        performance counters and annotates them in the trace.
            * The events are logged under the `cuda_profiler_range` category.
            * Counter values are logged in the `args` section of the trace.

        This API can investigate performance measurements per kernel and
        associate them to operators that the kernel belongs to. A single kernel
        can map to multiple levels of operators (as operators can be nested).
        To represent this we basically provide a list column called `op_stack`.
        For further convenience we add the top and bottom level operator column
        as well.

        Args:
            ranks (List[int]): List of ranks on which to run the analysis. Default = [0].
        Returns:
            List[pd.DataFrame]
                A list of dataframes, one per rank, containing kernel name,
                op_stack (operator stack), top and bottom level op, and columns
                for individual performance counters.
        """
        return CuptiCounterAnalysis.get_counter_data_with_operators(self.t, ranks)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Callable, List, Optional, TYPE_CHECKING

import pandas as pd

from hta.analyzers.straggler import (
    extract_iteration_info,
    find_stragglers_with_late_start_comm_kernels,
)
from hta.configs.config import logger

# import statement used without the "if TYPE_CHECKING" guard will cause a circular
# dependency with trace_analysis.py causing mypy to fail and should not be removed.
if TYPE_CHECKING:
    from hta.common.trace import Trace


class StragglerAnalysis:
    def __init__(self):
        pass

    @classmethod
    def get_profiler_steps(cls, t: "Trace") -> List[int]:
        """
        Profiler steps implementation. Returns the list of profiler steps.
        """
        return sorted(
            [i for i in extract_iteration_info(t)["iter"].unique() if i != -1]
        )

    @classmethod
    def get_potential_stragglers(
        cls,
        t: "Trace",
        profiler_steps: Optional[List[int]] = None,
        num_candidates: int = 2,
        visualize: bool = False,
        straggler_identification_impl: Callable[
            ..., pd.Series
        ] = find_stragglers_with_late_start_comm_kernels,
    ) -> List[int]:
        """
        Straggler analysis implementation. See `get_potential_stragglers` in `trace_analysis.py` for details.
        """
        if num_candidates < 1:
            num_candidates = 1

        available_profiler_steps = cls.get_profiler_steps(t)
        if profiler_steps is None:
            valid_profiler_steps = available_profiler_steps
        else:
            valid_profiler_steps = [
                i for i in profiler_steps if i in available_profiler_steps
            ]
        if len(valid_profiler_steps) == 0:
            raise ValueError(
                f"invalid value for argument: profiler_steps={profiler_steps}; available profiler steps={available_profiler_steps}"
            )

        ranks = list(t.get_all_traces().keys())
        df_all = pd.concat(
            [t.get_trace(r) for r in ranks], axis=0, keys=ranks, names=["rank", "idx"]
        ).reset_index()

        df_selected_profiler_steps = df_all.loc[
            df_all["iteration"].isin(valid_profiler_steps)
        ]

        straggler_counts = straggler_identification_impl(
            df_selected_profiler_steps, t.symbol_table, num_candidates, visualize
        )
        stragglers = straggler_counts.sort_values(ascending=False)[
            :num_candidates
        ].index.tolist()

        if len(stragglers) > 1:
            logger.debug(f"found ranks {stragglers} are potential stragglers.")
        else:
            logger.debug(f"found rank {stragglers} is a potential straggler.")

        return stragglers



# Copyright (c) Meta Platforms, Inc. and affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import unittest
from collections import namedtuple
from pathlib import Path
from typing import List
from unittest.mock import patch

import hta
from hta.analyzers.cupti_counter_analysis import CUDA_SASS_INSTRUCTION_COUNTER_FLOPS
from hta.common.trace import PHASE_COUNTER
from hta.trace_analysis import TimeSeriesTypes, TraceAnalysis


class TraceAnalysisTestCase(unittest.TestCase):
    vision_transformer_t: TraceAnalysis
    inference_t: TraceAnalysis
    df_index_resolver_t: TraceAnalysis
    rank_non_gpu_t: TraceAnalysis

    @classmethod
    def setUpClass(cls):
        super(TraceAnalysisTestCase, cls).setUpClass()
        vision_transformer_trace_dir: str = "tests/data/vision_transformer"
        inference_trace_dir: str = "tests/data/inference_single_rank"
        df_index_resolver_trace_dir: str = "tests/data/df_index_resolver"
        rank_non_gpu_trace_dir: str = "tests/data/rank_non_gpu/"
        cls.vision_transformer_t = TraceAnalysis(trace_dir=vision_transformer_trace_dir)
        cls.inference_t = TraceAnalysis(trace_dir=inference_trace_dir)
        cls.df_index_resolver_t = TraceAnalysis(trace_dir=df_index_resolver_trace_dir)
        cls.rank_non_gpu_t = TraceAnalysis(trace_dir=rank_non_gpu_trace_dir)

    def setUp(self):
        self.overlaid_trace_dir = "tests/data"
        self.overlaid_trace_file = os.path.join(
            str(Path(self.overlaid_trace_dir)), "overlaid_rank-0.json.gz"
        )

    @patch.object(hta.common.trace.Trace, "write_raw_trace")
    def test_frequent_cuda_kernel_sequences(self, mock_write_trace):
        frequent_patterns_dfs = (
            self.vision_transformer_t.get_frequent_cuda_kernel_sequences(
                operator_name="aten::linear",
                output_dir=self.overlaid_trace_dir,
                visualize=False,
                compress_other_kernels=True,
            )
        )
        self.assertIn(
            "|".join(
                [
                    "aten::linear",
                    "Memset (Device)",
                    "volta_fp16_s884gemm_fp16_128x128_ldg8_f2f_stages_32x1_tn",
                    "void at::native::elementwise_kernel",
                ]
            ),
            frequent_patterns_dfs.iloc[2]["pattern"],
        )
        self.assertEqual(frequent_patterns_dfs.iloc[2]["count"], 48)
        self.assertEqual(
            frequent_patterns_dfs.iloc[2]["GPU kernel duration (us)"], 11300
        )
        self.assertEqual(frequent_patterns_dfs.iloc[2]["CPU op duration (us)"], 9652)
        mock_write_trace.assert_called_once()
        trace_output_filename, _ = mock_write_trace.call_args.args
        self.assertEqual(trace_output_filename, self.overlaid_trace_file)

    def test_no_frequent_cuda_kernel_sequences_found(self):
        frequent_patterns_dfs = (
            self.df_index_resolver_t.get_frequent_cuda_kernel_sequences(
                operator_name="aten::clone",
                output_dir=self.overlaid_trace_dir,
                rank=1,
                visualize=False,
                compress_other_kernels=True,
            )
        )
        self.assertTrue(frequent_patterns_dfs.empty)

    def test_get_cuda_kernel_launch_stats_training_multiple_ranks(self):
        dataframe_dict = self.vision_transformer_t.get_cuda_kernel_launch_stats(
            ranks=[1, 7], visualize=False
        )
        rank_1_df, rank_7_df = dataframe_dict[1], dataframe_dict[7]
        row1 = rank_1_df[rank_1_df["correlation"] == 373234]
        row2 = rank_7_df[rank_7_df["correlation"] == 327327]

        self.assertEqual(row1["cpu_duration"].item(), 16)
        self.assertEqual(row1["gpu_duration"].item(), 2394)
        self.assertEqual(row1["launch_delay"].item(), 16491)
        self.assertEqual(row2["cpu_duration"].item(), 21)
        self.assertEqual(row2["gpu_duration"].item(), 94)
        self.assertEqual(row2["launch_delay"].item(), 41)

    def test_get_cuda_kernel_launch_stats_inference_single_rank(self):
        dataframe_list = self.inference_t.get_cuda_kernel_launch_stats(visualize=False)
        rank_0_df = dataframe_list[0]
        row = rank_0_df[rank_0_df["correlation"] == 684573]

        self.assertEqual(row["cpu_duration"].item(), 9)
        self.assertEqual(row["gpu_duration"].item(), 3)
        self.assertEqual(row["launch_delay"].item(), 20)

    def test_get_profiler_steps(self):
        results = self.vision_transformer_t.get_profiler_steps()
        expected = [15, 16, 17, 18]
        self.assertListEqual(results, expected)

    def test_get_potential_stragglers(self):
        TCase = namedtuple(
            "TCase", ["profiler_steps", "num_candidates", "expected_results"]
        )
        p_steps = self.vision_transformer_t.get_profiler_steps()
        test_cases: List[TCase] = [
            TCase(p_steps[:1], -1, [7]),
            TCase(p_steps[:1], 2, [6, 7]),
            TCase(p_steps, 2, [0, 1]),
        ]

        for tc in test_cases:
            got_stragglers = sorted(
                self.vision_transformer_t.get_potential_stragglers(
                    profiler_steps=tc.profiler_steps, num_candidates=tc.num_candidates
                )
            )
            self.assertListEqual(got_stragglers, tc.expected_results)

    def test_comm_comp_overlap(self):
        comm_comp_overlap = self.vision_transformer_t.get_comm_comp_overlap(
            visualize=False
        )
        self.assertAlmostEqual(
            comm_comp_overlap.iloc[0]["comp_comm_overlap_pctg"],
            round((240322 * 100) / 1091957, 3),
            delta=0.01,
        )
        self.assertAlmostEqual(
            comm_comp_overlap.iloc[1]["comp_comm_overlap_pctg"],
            round((262584 * 100) / 1233755, 3),
            delta=0.01,
        )

    def test_temporal_breakdown(self):
        idle_time = self.vision_transformer_t.get_temporal_breakdown(visualize=False)
        self.assertAlmostEqual(
            idle_time.iloc[0]["idle_time_pctg"],
            round((552069 * 100) / 2033570, 3),
            delta=0.01,
        )
        self.assertAlmostEqual(
            idle_time.iloc[0]["compute_time_pctg"],
            round((596651 * 100 / 2033570), 3),
            delta=0.01,
        )
        self.assertAlmostEqual(
            idle_time.iloc[0]["non_compute_time_pctg"],
            round(884850 * 100 / 2033570, 3),
            delta=0.01,
        )
        self.assertAlmostEqual(
            idle_time.iloc[1]["idle_time_pctg"],
            round(431771 * 100 / 2032757, 3),
            delta=0.01,
        )
        self.assertAlmostEqual(
            idle_time.iloc[1]["compute_time_pctg"],
            round(596759 * 100 / 2032757, 3),
            delta=0.01,
        )
        self.assertAlmostEqual(
            idle_time.iloc[1]["non_compute_time_pctg"],
            round(1004227 * 100 / 2032757, 3),
            delta=0.01,
        )

    def test_get_gpu_kernel_breakdown(self):
        (
            kernel_type_breakdown,
            kernel_breakdown,
        ) = self.vision_transformer_t.get_gpu_kernel_breakdown(
            visualize=False, include_memory_kernels=True
        )

        self.assertEqual(kernel_type_breakdown.iloc[0]["kernel_type"], "COMMUNICATION")
        self.assertEqual(kernel_type_breakdown.iloc[0]["sum"], 8040285)
        self.assertEqual(kernel_breakdown.iloc[0]["kernel_type"], "COMMUNICATION")
        self.assertEqual(kernel_breakdown.iloc[0]["sum (us)"], 627683)
        self.assertEqual(kernel_breakdown.iloc[151]["kernel_type"], "MEMORY")
        self.assertEqual(kernel_breakdown.iloc[151]["sum (us)"], 1064)

    def test_get_queue_length_summary(self):
        qd_summary = self.vision_transformer_t.get_queue_length_summary(ranks=[0])
        streams = qd_summary.index.to_list()
        self.assertEqual(streams, list(zip([0] * 6, [7, 20, 24, 26, 28, 30])))

        stream7_stats = qd_summary.loc[0, 7]["queue_length"].to_dict()
        expected_stats = {
            "count": 17160.0,
            "mean": 61.043473193,
            "std": 92.33450047722,
            "min": 0.0,
            "25%": 1.0,
            "50%": 9.0,
            "75%": 87.0,
            "max": 403.0,
        }
        for key, expval in expected_stats.items():
            self.assertAlmostEqual(
                stream7_stats[key], expval, msg=f"Stream 7 stats mismatch key={key}"
            )

    @patch.object(hta.common.trace.Trace, "write_raw_trace")
    def test_generate_trace_with_counters(self, mock_write_trace):
        self.vision_transformer_t.generate_trace_with_counters(
            time_series=TimeSeriesTypes.QUEUE_LENGTH | TimeSeriesTypes.MEMCPY_BANDWIDTH
        )
        mock_write_trace.assert_called_once()
        # change to kwargs if you use kwargs while calling write_raw_trace
        trace_filename, trace_json = mock_write_trace.call_args.args
        self.assertTrue("with_counters" in trace_filename)

        counter_events = [
            ev for ev in trace_json["traceEvents"] if ev["ph"] == PHASE_COUNTER
        ]
        print(f"Trace has {len(counter_events)} counter events")
        self.assertGreaterEqual(len(counter_events), 21000)

        counter_names = {ev["name"] for ev in counter_events}
        self.assertEqual(
            counter_names,
            {"Queue Length", "Memcpy DtoH", "Memcpy HtoD", "Memcpy DtoD", "Memset"},
        )

        membw_ts = self.vision_transformer_t.get_memory_bw_time_series()[0]
        self.assertEqual(len(membw_ts[membw_ts.memory_bw_gbps < 0]), 0)

        queue_len_ts = self.vision_transformer_t.get_queue_length_time_series()[0]
        self.assertEqual(len(queue_len_ts[queue_len_ts.queue_length < 0]), 0)

        mem_bw_summary_df = self.vision_transformer_t.get_memory_bw_summary(
            ranks=[0, 2]
        )
        # 2 ranks x 4 types of memcpy/memset
        self.assertEqual(len(mem_bw_summary_df), 8)

        queue_len_summary_df = self.vision_transformer_t.get_queue_length_summary(
            ranks=[0, 2]
        )
        # 2 ranks x 6 streams
        self.assertEqual(len(queue_len_summary_df), 12)

        # Test traces without GPU kernels, these should return empty dicts or dataframes
        queue_len_ts_dict = self.rank_non_gpu_t.get_queue_length_time_series()
        self.assertEqual(len(queue_len_ts_dict), 0)

        queue_len_summary_df = self.rank_non_gpu_t.get_queue_length_summary(ranks=[0])
        self.assertIsNone(queue_len_summary_df)

        mem_bw_summary_df = self.rank_non_gpu_t.get_memory_bw_summary(ranks=[0])
        self.assertIsNone(mem_bw_summary_df)

    def test_get_idle_time_breakdown(self):
        (
            idle_time_df,
            idle_interval_df,
        ) = self.vision_transformer_t.get_idle_time_breakdown(
            ranks=[0, 1], visualize=False, show_idle_interval_stats=True
        )
        ranks = idle_time_df[
            "rank"
        ].unique()  # cannot use dataframe.rank as rank() is a utility too
        streams = idle_time_df.stream.unique()
        idle_categories = idle_time_df.idle_category.unique()

        self.assertEqual(set(ranks), {0, 1})
        self.assertEqual(set(streams), {7, 20, 24, 26, 28, 30})
        self.assertEqual(set(idle_categories), {"host_wait", "kernel_wait", "other"})

        # Ratios sum up to 1.0, 6 streams x 2 ranks = 12.0
        self.assertAlmostEqual(idle_time_df.idle_time_ratio.sum(), 12.0)

        stream7_stats = idle_time_df[idle_time_df.stream == 7].iloc[0].to_dict()
        expected_stats = {
            "idle_category": "host_wait",
            "idle_time": 1000581.0,
            "stream": 7,
            "idle_time_ratio": 0.71,
            "rank": 0,
        }
        for key, expval in expected_stats.items():
            self.assertAlmostEqual(
                stream7_stats[key],
                expval,
                msg=f"Stream 7 idle stats mismatch key={key}",
            )

    def test_get_cupti_counter_data_with_operators(self):
        # regular trace should return empty list since it will not have cuda_profiler_range events
        self.assertEqual(self.inference_t.get_cupti_counter_data_with_operators(), [])

        cupti_profiler_trace_dir: str = "tests/data/cupti_profiler/"
        cupti_profiler_t = TraceAnalysis(trace_dir=cupti_profiler_trace_dir)
        counters_df = cupti_profiler_t.get_cupti_counter_data_with_operators()[0]

        self.assertEqual(len(counters_df), 77)

        # Example trace has CUPTI SASS FLOPS instruction counters
        counter_names = set(CUDA_SASS_INSTRUCTION_COUNTER_FLOPS.keys())
        self.assertEqual(
            set(counters_df.columns.unique()) & counter_names, counter_names
        )

        # Pick 5th kernel that executed.
        test_row = counters_df.sort_values(axis=0, by="ts").iloc[5].to_dict()

        self.assertEqual(test_row["cat"], "cuda_profiler_range")
        self.assertTrue("fft2d_r2c_32x32" in test_row["name"])

        self.assertEqual(
            test_row["smsp__sass_thread_inst_executed_op_fadd_pred_on.sum"], 86114304
        )
        self.assertEqual(test_row["top_level_op"], "aten::conv2d")
        self.assertEqual(test_row["bottom_level_op"], "aten::_convolution")


if __name__ == "__main__":  # pragma: no cover
    unittest.main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import unittest

import pandas as pd

from hta.common.trace import parse_trace_dict, Trace


class TraceParseTestCase(unittest.TestCase):
    vision_transformer_t: Trace
    vision_transformer_raw_df: pd.DataFrame
    inference_t: Trace
    inference_raw_df: pd.DataFrame

    @classmethod
    def setUpClass(cls):
        super(TraceParseTestCase, cls).setUpClass()
        vision_transformer_trace_dir: str = "tests/data/vision_transformer"
        inference_trace_dir: str = "tests/data/inference_single_rank"
        vision_transformer_rank_0_file: str = "rank-0.json.gz"
        inference_rank_0_file: str = "inference_rank_0.json.gz"
        max_ranks = 8

        # Trace parser for vision transformer
        cls.vision_transformer_t: Trace = Trace(trace_dir=vision_transformer_trace_dir)
        cls.vision_transformer_t.parse_traces(
            max_ranks=max_ranks, use_multiprocessing=True
        )
        cls.vision_transformer_raw_df = cls.prepare_ground_truth_df(
            vision_transformer_trace_dir, vision_transformer_rank_0_file
        )
        # Trace parser for inference
        cls.inference_t: Trace = Trace(trace_dir=inference_trace_dir)
        cls.inference_t.parse_traces(max_ranks=max_ranks, use_multiprocessing=True)
        cls.inference_raw_df = cls.prepare_ground_truth_df(
            inference_trace_dir, inference_rank_0_file
        )

    @classmethod
    def prepare_ground_truth_df(cls, trace_dir, rank_0_file) -> pd.DataFrame:
        df = pd.DataFrame(
            parse_trace_dict(os.path.join(trace_dir, rank_0_file))["traceEvents"]
        )
        df.dropna(axis=0, subset=["dur", "cat"], inplace=True)
        df.drop(df[df["cat"] == "Trace"].index, inplace=True)
        return df

    def setUp(self) -> None:
        self.traces = [self.vision_transformer_t, self.inference_t]
        self.raw_dfs = [self.vision_transformer_raw_df, self.inference_raw_df]
        self.total_ranks = [8, 1]

    def test_trace_load(self) -> None:
        # run tests for each collection of traces
        for t, raw_df, total_ranks in zip(self.traces, self.raw_dfs, self.total_ranks):
            # test raw trace after parsing
            self.assertEqual(len(t.traces), total_ranks)

            sym_id_map = t.symbol_table.get_sym_id_map()
            sym_table = t.symbol_table.get_sym_table()
            rank_0_df_name_id = t.traces[0]["name"]
            rank_0_df_name = t.traces[0]["name"].apply(lambda x: sym_table[x])

            ground_truth_name = raw_df["name"]
            ground_truth_name_id = raw_df["name"].apply(lambda x: sym_id_map[x])

            self.assertListEqual(
                rank_0_df_name_id.to_list(), ground_truth_name_id.to_list()
            )
            self.assertListEqual(rank_0_df_name.to_list(), ground_truth_name.to_list())

            # test aligned and filtered trace
            t.align_and_filter_trace()

            raw_profiler_steps = raw_df["name"].str.contains("ProfilerStep").sum()
            sym_id_map = t.symbol_table.get_sym_id_map()
            profiler_steps = [v for k, v in sym_id_map.items() if "ProfilerStep" in k]
            filtered_profiler_steps = t.traces[0]["name"].isin(profiler_steps).sum()

            self.assertEqual(
                filtered_profiler_steps + int(raw_profiler_steps > 1),
                raw_profiler_steps,
            )
            self.assertLessEqual(len(t.traces[0]), len(raw_df))
            self.assertGreaterEqual(t.traces[0]["ts"].min(), 0)

    def test_trace_iteration(self) -> None:
        # run tests for each collection of traces
        for t in self.traces:
            df = t.traces[0]
            sym_id_map = t.symbol_table.get_sym_id_map()
            iterations = {
                f"ProfilerStep#{i}" for i in set(df["iteration"].unique()) if i != -1
            }

            valid_gpu_kernels = df.loc[
                df["stream"].gt(0) & df["index_correlation"].gt(0)
            ]
            correlated_cpu_ops = df.loc[
                df.loc[valid_gpu_kernels.index, "index_correlation"]
            ]
            gpu_kernels_per_iteration = (
                valid_gpu_kernels.groupby("iteration")["index"].agg("count").to_dict()
            )
            correlated_cpu_ops_per_iteration = (
                correlated_cpu_ops.groupby("iteration")["index"].agg("count").to_dict()
            )

            self.assertTrue("iteration" in df.columns)
            self.assertTrue(all(i in sym_id_map for i in iterations))
            self.assertDictEqual(
                gpu_kernels_per_iteration, correlated_cpu_ops_per_iteration
            )


if __name__ == "__main__":  # pragma: no cover
    unittest.main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import gzip
import json
import os
import re
from typing import Any, Dict, Tuple

from hta.configs.config import logger


def create_rank_to_trace_dict(trace_dir: str) -> Tuple[bool, Dict]:
    """
    Create a rank -> trace_filename map for traces located within the directory <trace_path>

    Args:
        trace_dir (str) : the path to the directory where the traces are located.

    Returns:
        (success: bool, rank_trace_map: dict) : a tuple indicating whether the operation succeeds and the path
         to the map file being created.
            success (bool) : True if the file is created successfully; False otherwise.
            map_file_path (dict) : a dict with rank as key and trace_filename as the value
    """

    if not (os.path.exists(trace_dir) and os.access(trace_dir, os.R_OK)):
        logger.error(f"trace_path {trace_dir} doesn't exist or is not readable")
        return False, {}

    file_list = [
        fn for fn in os.listdir(trace_dir) if fn.endswith(".gz") or fn.endswith(".json")
    ]
    if len(file_list) == 0:
        logger.warning(f"No trace file is found in {trace_dir}")
        return False, {}

    rank_to_trace_dict: Dict[int, str] = {}
    for file in file_list:
        file_path = os.path.join(trace_dir, file)

        with gzip.open(file_path, "rb") if file.endswith("gz") else open(
            file_path, "r"
        ) as f:
            for line in f:
                data = line.decode() if isinstance(line, bytes) else line
                if "rank" in data:
                    break

            # match is like "rank": 6,
            match = re.search(r'"rank": \d+', data)
            if match:
                rank = int(match.group().split(": ")[1])
                if rank in rank_to_trace_dict:
                    logger.error(
                        f"File {rank_to_trace_dict[rank]} and file {file_path} has the same rank. Will use {file_path} as the path to rank: {rank}."
                    )
                rank_to_trace_dict[int(rank)] = file_path
            else:
                logger.error(
                    "If the trace file does not have the rank specified in it, then add the following snippet "
                    'key to the json files to use HTA; "distributedInfo": {"rank": 0}. If there are multiple '
                    "traces files, then each file should have a unique rank value."
                )

    return True, rank_to_trace_dict


def get_trace_files(trace_path: str) -> Dict[int, str]:
    """
    Get the rank to trace map from traces in the directory <trace_path>.

    Args:
        trace_path (str) : the path to the directory where the traces are located.

    Returns:
        rank_trace_file_map (Dict[int, str]) : a dictionary with rank as key and trace filename as value.
        Only valid rank_trace_file_map entry will be included in the dictionary. A rank_trace_file_map entry
        is valid only when  (1) the rank is a positive integer, and (2) the trace file exists and is readable.
    """
    rank_to_trace_dict: Dict[int, str] = {}
    if not os.path.exists(trace_path):
        logger.warning(f"{trace_path} is not a valid path")
    else:
        ok, rank_to_trace_dict = create_rank_to_trace_dict(trace_path)
        if not ok:
            logger.warning("failed to create rank to trace map")
            return {}

    if len(rank_to_trace_dict) == 0:
        logger.warning("There is no item in the rank to trace file map.")
    else:
        logger.info(f"Rank to trace file map:\n{rank_to_trace_dict}")
    return rank_to_trace_dict


def read_trace(file_path: str) -> Dict[str, Any]:
    """
    Read the trace from a file.

    Args:
        file_path (str): a trace file with ".gz" or "json" postfix.

    Returns:
        trace_data (Dict[str, Any]): the raw trace data
    """
    trace_data: Dict[str, Any] = {}
    if file_path.endswith(".gz"):
        with gzip.open(file_path, "rb") as fh:
            trace_data = json.loads(fh.read())
    elif file_path.endswith(".json"):
        with open(file_path, "r") as fh2:
            trace_data = json.loads(fh2.read())
    else:
        raise ValueError(f"trace_file ({file_path}) must end with '.gz' or 'json'")
    return trace_data


def write_trace(trace_data: Dict[str, Any], file_path: str) -> None:
    """
    Write a trace into a file.

    Args:
        trace_data (Dict[str, Any]): the trace data
        file_path (str): a trace file with ".gz" or "json" postfix.

    Returns:
        None
    """
    base_dir = os.path.dirname(file_path)
    if not os.path.exists(base_dir):
        os.makedirs(base_dir)

    if file_path.endswith(".gz"):
        json_str = json.dumps(trace_data)
        json_bytes = json_str.encode("utf-8")
        with gzip.open(file_path, "wb") as fp:
            fp.write(json_bytes)
    else:
        with open(file_path, "w+") as fp:
            fp.write(json.dumps(trace_data, indent=2))


def update_trace_rank(file_path: str, rank: int) -> None:
    """
    Add or update a rank metadata entry in a trace file.

    Args:
        file_path (str): a trace file.
        rank (int): the rank assigned to the trace file.

    Returns:
        None
    """

    def _add_rank_meta(trace_data: Dict[str, Any], rank: int) -> None:
        if "distributedInfo" in trace_data:
            trace_data["distributedInfo"]["rank"] = rank
        else:
            trace_data["distributedInfo"] = {"rank": rank}

    trace_data = read_trace(file_path)
    _add_rank_meta(trace_data, rank)
    write_trace(trace_data, file_path)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Tuple, Union

import pandas as pd
import plotly.express as px

from hta.common.call_stack import CallGraph
from hta.common.trace import Trace
from hta.configs.config import logger


class CudaKernelAnalysis:
    def __init__(self):
        pass

    @classmethod
    def get_frequent_cuda_kernel_sequences(
        cls,
        t: "Trace",
        operator_name: str,
        output_dir: str,
        min_pattern_len: int = 3,
        rank: int = 0,
        top_k: int = 5,
        visualize: bool = True,
        compress_other_kernels: bool = True,
    ) -> pd.DataFrame:
        """
        Frequent CUDA kernel sequences implementation. See `get_frequent_cuda_kernel_sequences` in `trace_analysis.py`
        for details.
        """
        if not operator_name:
            logger.error(
                "operator_name must be a non-empty string and a valid operator name in the trace file."
            )
            return pd.DataFrame()

        if not output_dir:
            logger.error(
                "output_dir must be a non-empty string and a valid folder name."
            )
            return pd.DataFrame()

        # check output_dir is a valid path and a directory
        path = Path(output_dir)
        if not path.exists():
            logger.error(f"The path {str(path)} does not exist.")
            return pd.DataFrame()
        if not path.is_dir():
            logger.error(f"{output_dir} is not a directory.")
            return pd.DataFrame()

        sym_index = t.symbol_table.get_sym_id_map()
        cg = CallGraph(t, ranks=[rank])
        trace_df = t.get_trace(rank)

        cpu_kernels = trace_df[trace_df["stream"].eq(-1)].copy()
        gpu_kernels = trace_df[trace_df["stream"].ne(-1)].copy()

        # get all the CPU operators which contain operator_name in their names
        candidate_root_idx = [
            idx for name, idx in sym_index.items() if operator_name in name
        ]
        candidate_nodes = cpu_kernels[cpu_kernels["name"].isin(candidate_root_idx)]
        # to avoid double-counting when the same CPU operators appear multiple times in the call graph
        # hierarchy, only start the search from the shallowest CPU operators
        min_depth = candidate_nodes["depth"].min()
        root_nodes = candidate_nodes[candidate_nodes["depth"].eq(min_depth)][
            ["index", "dur"]
        ]

        pattern_counts: Dict[Tuple[int, ...], int] = defaultdict(int)
        # duration of the pattern: [GPU kernel duration, CPU op duration]
        pattern_durations: Dict[Tuple[int, ...], List[int]] = defaultdict(
            lambda: [0, 0]
        )
        # record the indices of frequent patterns.
        pattern_occurrences: Dict[Tuple[int, ...], Set[int]] = defaultdict(set)

        for cs in cg.call_stacks:
            for _, root_node, root_dur in root_nodes.itertuples():
                if root_node in cs.get_nodes():
                    # find all cuda kernels launched by the root_node by joining GPU kernels with leaf nodes.
                    cuda_kernels = gpu_kernels.merge(
                        cpu_kernels[
                            cpu_kernels["index"].isin(cs.get_leaf_nodes(root_node))
                        ][["correlation"]],
                        on="correlation",
                        how="inner",
                    ).sort_values(by="ts")
                    # discard the pattern if it is too short
                    if len(cuda_kernels) < min_pattern_len:
                        continue
                    cpu_operator = cpu_kernels[cpu_kernels["index"].eq(root_node)]
                    pattern: Tuple[int, ...] = tuple(
                        pd.concat([cpu_operator["name"], cuda_kernels["name"]])
                    )
                    pattern_counts[pattern] += 1
                    # update the duration for GPU kernels in the pattern
                    pattern_durations[pattern][0] += cuda_kernels["dur"].sum()
                    # update the duration for CPU operators in the pattern
                    pattern_durations[pattern][1] += root_dur
                    for p in cs.get_paths_to_leaves(root_node):
                        pattern_occurrences[pattern].update(p)
                    pattern_occurrences[pattern].update(cuda_kernels["index"])

        if not pattern_counts:
            logger.error(
                "operator_name not found in the trace file, or no frequent cuda kernel sequences found."
            )
            return pd.DataFrame()

        return cls._generate_frequent_pattern_results(
            t,
            pattern_counts,
            pattern_durations,
            pattern_occurrences,
            rank,
            top_k,
            visualize,
            output_dir,
            compress_other_kernels,
        )

    @classmethod
    def _generate_frequent_pattern_results(
        cls,
        t: "Trace",
        pattern_counts: Dict[Tuple[int, ...], int],
        pattern_durations: Dict[Tuple[int, ...], List[int]],
        pattern_occurrences: Dict[Tuple[int, ...], Set[int]],
        rank: int,
        top_k: int,
        visualize: bool,
        output_dir: str,
        compress_other_kernels: bool = True,
    ) -> pd.DataFrame:
        """
        Post-process the frequent patterns:
        - generate the final dataframe;
        - overlay the frequent pattern results on the raw trace file;
        - create inline visualization.

        Args:
            pattern_counts (Dict[Tuple[int, ...], int]): frequent patterns and their counts
            pattern_durations (Dict[Tuple[int, ...], List[int]]): frequent patterns and their total GPU kernel
                                                                  and CPU op duration
            pattern_occurrences (Dict[Tuple[int, ...], Set[int]]): frequent patterns and where they happen
            rank (int): rank number on which the analysis should be performed on
            top_k (int): top_k patterns in terms of frequency to be visualized and overlaid
            visualize (bool): whether to show the histogram of top_k frequent patterns inline
            output_dir (str): path to the folder containing the new trace file with overlaid top k
                              frequent patterns
            compress_other_kernels (bool): should the names and args for other kernels not belonging to
                                           any frequent patterns be compressed to save memory in the overlaid
                                           trace file
        Returns:
            patterns_df (pd.DataFrame): a dataframe with all patterns and their frequencies
        """
        output_file = os.path.join(
            str(Path(output_dir)), "overlaid_" + t.trace_files[rank].split("/")[-1]
        )

        sym_table = t.symbol_table.get_sym_table()
        patterns_result: Dict[str, List[Union[int, str, Set[int]]]] = defaultdict(list)
        for pattern, count in pattern_counts.items():
            patterns_result["pattern"].append("|".join(sym_table[x] for x in pattern))
            patterns_result["count"].append(count)
            patterns_result["GPU kernel duration (us)"].append(
                pattern_durations[pattern][0]
            )
            patterns_result["CPU op duration (us)"].append(
                pattern_durations[pattern][1]
            )
            patterns_result["pattern_indices"].append(pattern_occurrences[pattern])

        patterns_df = pd.DataFrame(patterns_result).sort_values(
            by=["count", "pattern"], ascending=[False, True], ignore_index=True
        )
        if top_k > len(patterns_df):
            top_k = len(patterns_df)
            logger.info(
                "The top_k argument value exceeds the number of patterns. Displaying all patterns."
            )

        overlaid_trace = cls._overlay_frequent_patterns_with_trace(
            t, patterns_df[:top_k], rank, compress_other_kernels
        )
        t.write_raw_trace(output_file, overlaid_trace)
        logger.info(
            f"Overlaid trace file for rank {rank} has been generated at {output_file}."
        )
        logger.info(
            'View the generated trace file using Chrome Tracing and search for "Patterns" to highlight the frequent patterns.'
        )

        if visualize:  # pragma: no cover
            vis_df = patterns_df[:top_k].copy()
            # show the pattern in multiple lines in visualization
            vis_df["pattern"] = vis_df["pattern"].str.replace("|", "<br>", regex=False)
            fig = px.bar(
                vis_df,
                x=vis_df.index,
                y="count",
                hover_data=["pattern", "count"],
                title=f"Top {top_k} frequent patterns for rank {rank}",
                labels={"x": "Pattern Index", "count": "Pattern Count"},
            )
            fig.update_xaxes(type="category")
            fig.show()
        return patterns_df.drop("pattern_indices", axis=1)

    @classmethod
    def _overlay_frequent_patterns_with_trace(
        cls,
        t: "Trace",
        patterns_df: pd.DataFrame,
        rank: int = 0,
        compress_other_kernels: bool = True,
    ) -> Dict[str, Any]:
        """
        Overlay the identified frequent patterns on top of the trace file for visualization.

        Args:
            patterns_df (pd.DataFrame): a dataframe with all patterns and their frequencies
            rank (int): the rank number on which the analysis was performed on
            compress_other_kernels (bool): should the names and args for other kernels not belonging to
                                           any frequent patterns be compressed to save memory in the overlaid
                                           trace file
        Returns:
            raw_trace_content (Dict[str, Any]): the dictionary form of the overlaid trace file that can be dumped
        """
        raw_trace_content = t.get_raw_trace_for_one_rank(rank=rank)
        raw_trace_df = pd.DataFrame(raw_trace_content["traceEvents"]).reset_index()
        raw_trace_df["index"] = pd.to_numeric(raw_trace_df["index"], downcast="integer")
        sym_id_map = t.symbol_table.get_sym_id_map()

        # get a counter of patterns that each CPU operator/GPU kernel is in.
        top_patterns_df = (
            patterns_df.explode("pattern_indices")
            .groupby("pattern_indices", as_index=False)
            .apply(
                lambda g: pd.Series(
                    {
                        "active_patterns": {
                            r["pattern"]: r["count"] for _, r in g.iterrows()
                        }
                    }
                )
            )
        )
        # join the pattern counter with the original events on their indices
        merged_df = raw_trace_df.merge(
            top_patterns_df, left_on="index", right_on="pattern_indices", how="left"
        )

        def _add_patterns_to_args(row, compress_other_kernels):
            args = row["args"]
            if pd.isna(row["pattern_indices"]):
                # only drop complete events to ensure important information is retained
                return {} if compress_other_kernels and row["ph"] == "X" else args
            # add the frequent patterns to the args field
            if pd.isna(args):
                args = {}
            args["Patterns"] = row["active_patterns"]
            return args

        def _compress_kernel_names(row):
            # do not compress events that are part of a frequent pattern, or we don't know how to compress
            if row["args"] or row["name"] not in sym_id_map:
                return row["name"]
            # otherwise return the symbol id created during trace parsing of that event
            return str(sym_id_map[row["name"]])

        # add the frequent patterns information to the args field so that it is searchable
        merged_df["args"] = merged_df.apply(
            lambda r: _add_patterns_to_args(r, compress_other_kernels), axis=1
        )

        if compress_other_kernels:
            # compress kernel/operator names
            merged_df["name"] = merged_df.apply(_compress_kernel_names, axis=1)
            # add the symbol id -> event name mapping information to the PyTorch Profiler event for reference
            profiler_event = raw_trace_df[
                (raw_trace_df["name"].str.contains("PyTorch Profiler"))
                & (raw_trace_df["pid"].eq("Spans"))
            ]
            merged_df.loc[profiler_event.index, "args"] = [
                {str(i): name for name, i in sym_id_map.items()}
            ]
        # drop unused columns before writing back to the overlaid trace file
        merged_df.drop(
            ["index", "active_patterns", "pattern_indices"], axis=1, inplace=True
        )
        raw_trace_content["traceEvents"] = list(
            merged_df.apply(lambda row: row.dropna().to_dict(), axis=1)
        )
        return raw_trace_content

    @classmethod
    def visualize_cuda_kernel_launch_stats(
        cls, rank: int, df: pd.DataFrame, runtime_cutoff: int, launch_delay_cutoff: int
    ) -> None:  # pragma: no cover
        short_kernels = df[
            (df["cpu_duration"] <= runtime_cutoff)
            & (df["gpu_duration"] < df["cpu_duration"])
        ]
        runtime_outliers = df[df["cpu_duration"] > runtime_cutoff]
        launch_delay_outliers = df[df["launch_delay"] > launch_delay_cutoff]
        fig1 = px.histogram(
            short_kernels,
            x=short_kernels["cpu_duration"],
            title="Short GPU kernels (gpu op duration < cpu op duration) on rank %d"
            % rank,
            labels={
                "cpu_duration": "Cuda Runtime Event Duration (us)",
            },
        )
        fig1.show()

        fig2 = px.histogram(
            runtime_outliers,
            x=runtime_outliers["cpu_duration"],
            nbins=300,
            title="Runtime Event Duration Outliers (duration > %s us)  on rank %d"
            % (runtime_cutoff, rank),
            labels={
                "cpu_duration": "Cuda Runtime Event Duration (us)",
            },
        )
        fig2.show()

        fig3 = px.histogram(
            launch_delay_outliers,
            x=launch_delay_outliers["launch_delay"],
            nbins=300,
            title="Launch Delay Outliers (duration > %s us) on rank %d"
            % (launch_delay_cutoff, rank),
            labels={
                "launch_delay": "Launch Delay Duration (us)",
            },
        )
        fig3.show()

    @classmethod
    def cuda_kernel_launch_stats(
        cls,
        t: "Trace",
        ranks: Optional[List[int]] = None,
        runtime_cutoff: int = 50,
        launch_delay_cutoff: int = 100,
        include_memory_events: bool = True,
        visualize: bool = True,
    ) -> Dict[int, pd.DataFrame]:
        """
        CUDA kernel launch statistics implementation. For details see `get_cuda_kernel_launch_stats` in `trace_analysis.py`.
        """
        if ranks is None or ranks == []:
            ranks = [0]

        result_dict: Dict = {}
        sym_index = t.symbol_table.get_sym_id_map()

        for rank in ranks:
            # get trace for a rank
            trace_df: pd.DataFrame = t.get_trace(rank)

            # filter out events which have correlation value matching to
            # cudaLaunchKernel, cudaMemcpyAsync, cudaMemsetAsync
            cudaLaunchKernel_id = sym_index.get("cudaLaunchKernel", None)
            cudaMemcpyAsync_id = sym_index.get("cudaMemcpyAsync", None)
            cudaMemsetAsync_id = sym_index.get("cudaMemsetAsync", None)

            # get correlation id's of cudaLaunchKernel events
            cudaLaunchKernel_correlation_series: pd.Series = trace_df[
                trace_df["name"] == cudaLaunchKernel_id
            ].correlation

            # whether to use memory events - cudaMemsetAsync and cudaMemcpyAsync.
            if include_memory_events:
                memory_event_correlation_series: pd.Series = trace_df[
                    (trace_df["name"] == cudaMemsetAsync_id)
                    | (trace_df["name"] == cudaMemcpyAsync_id)
                ].correlation
                merged_series: pd.Series = pd.concat(
                    [
                        cudaLaunchKernel_correlation_series,
                        memory_event_correlation_series,
                    ]
                )
            else:
                merged_series = cudaLaunchKernel_correlation_series

            # filter cpu and gpu ops
            cpu_kernels = trace_df[trace_df["stream"].eq(-1)].copy()
            gpu_kernels = trace_df[trace_df["stream"].ne(-1)].copy()

            # get correlation id, duration, timestamp and name of cpu and gpu events
            cpu_kernels_filtered = cpu_kernels[
                (cpu_kernels["correlation"]).isin(merged_series)
            ][["correlation", "dur", "name", "ts"]]
            gpu_kernels_filtered = gpu_kernels[
                (gpu_kernels["correlation"]).isin(merged_series)
            ][["correlation", "dur", "name", "ts"]]

            # join the dataframes created above on correlation. This is required to calculate the launch delay
            # for each correlation id.
            joined_df = pd.merge(
                cpu_kernels_filtered,
                gpu_kernels_filtered,
                how="inner",
                on="correlation",
            )
            joined_df["launch_delay"] = (
                joined_df["ts_y"] - joined_df["ts_x"] + joined_df["dur_x"]
            )

            # rename columns and select the required columns from the final dataframe
            renamed_df = joined_df.rename(
                columns={"dur_x": "cpu_duration", "dur_y": "gpu_duration"}
            )
            events_df = renamed_df[
                ["correlation", "cpu_duration", "gpu_duration", "launch_delay"]
            ]

            if visualize:  # pragma: no cover
                cls.visualize_cuda_kernel_launch_stats(
                    rank, events_df, runtime_cutoff, launch_delay_cutoff
                )

            result_dict[rank] = events_df
        return result_dict



# Copyright (c) Meta Platforms, Inc. and affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import List

# Default Paths
DEFAULT_TRACE_DIR = "/tmp/trace"
DEFAULT_CONFIG_FILENAME: str = "trace_analyzer.json"

# Trace related
DF_SYMBOL_COLUMNS: List[str] = ["cat", "name"]

# Runtime configurations
IS_DEBUG_ENABLED: bool = True
MAX_NUM_PROCESSES: int = 32



# Copyright (c) Meta Platforms, Inc. and affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest

import pandas as pd

from hta.common.call_stack import CallStackGraph, CallStackIdentity, CallStackNode


class CallStackTestCase(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.df1 = pd.DataFrame(
            {
                "index": [0, 1, 2, 3, 4, 5, 6, 7, 8],
                "ts": [0, 1, 3, 3, 7, 8, 10, 12, 16],
                "dur": [10, 5, 2, 1, 3, 1, 11, 6, 3],
                "pid": [1, 1, 1, 1, 1, 1, 1, 1, 1],
                "tid": [2, 2, 2, 2, 2, 2, 2, 2, 2],
                "stream": [-1, -1, -1, -1, -1, -1, -1, -1, -1],
                "index_correlation": [-1, -1, -1, -1, -1, -1, -1, -1, -1],
            }
        )
        self.csi = CallStackIdentity(0, 1, 2)
        self.nodes = {
            -1: CallStackNode(parent=-1, depth=-1, children=[0, 6]),
            0: CallStackNode(parent=-1, depth=0, children=[1, 4]),
            1: CallStackNode(parent=0, depth=1, children=[2]),
            2: CallStackNode(parent=1, depth=2, children=[3]),
            3: CallStackNode(parent=2, depth=3, children=[]),
            4: CallStackNode(parent=0, depth=1, children=[5]),
            5: CallStackNode(parent=4, depth=2, children=[]),
            6: CallStackNode(parent=-1, depth=0, children=[7]),
            7: CallStackNode(parent=6, depth=1, children=[8]),
            8: CallStackNode(parent=7, depth=2, children=[]),
        }
        self.path_to_root_of_2 = [2, 1, 0, -1]
        self.path_to_root_of_3 = [3, 2, 1, 0, -1]
        self.path_to_root_of_5 = [5, 4, 0, -1]
        self.leaf_nodes_of_0 = [3, 5]
        self.leaf_nodes_of_4 = [5]
        self.leaf_nodes_of_5 = [5]
        self.paths_to_leaves_of_0 = [[0, 1, 2, 3], [0, 4, 5]]

    def test_construct_call_graph(self):
        csg = CallStackGraph(self.df1, self.csi)
        nodes = csg.get_nodes()
        self.assertDictEqual(nodes, self.nodes)

    def test_sort_events(self):
        index = [1, 2, 3, 4]
        start = [0, 0, 5, 5]
        dur = [10, 5, 1, 5]
        stream = [-1, -1, -1, -1]
        cor = [-1, -1, -1, -1]
        df = pd.DataFrame(
            {
                "index": index,
                "ts": start,
                "dur": dur,
                "stream": stream,
                "index_correlation": cor,
            }
        )
        nodes = {
            -1: CallStackNode(parent=-1, depth=-1, children=[1]),
            1: CallStackNode(parent=-1, depth=0, children=[2, 4]),
            2: CallStackNode(parent=1, depth=1, children=[]),
            4: CallStackNode(parent=1, depth=1, children=[3]),
            3: CallStackNode(parent=4, depth=2, children=[]),
        }
        csg = CallStackGraph(df, self.csi)
        self.assertDictEqual(nodes, csg.get_nodes())

    def test_get_path_to_root(self):
        csg = CallStackGraph(self.df1, self.csi)
        self.assertListEqual(csg.get_path_to_root(2), self.path_to_root_of_2)
        self.assertListEqual(csg.get_path_to_root(3), self.path_to_root_of_3)
        self.assertListEqual(csg.get_path_to_root(5), self.path_to_root_of_5)

    def test_get_leaf_nodes(self):
        csg = CallStackGraph(self.df1, self.csi)
        self.assertListEqual(csg.get_leaf_nodes(0), self.leaf_nodes_of_0)
        self.assertListEqual(csg.get_leaf_nodes(4), self.leaf_nodes_of_4)
        self.assertListEqual(csg.get_leaf_nodes(5), self.leaf_nodes_of_5)

    def test_get_paths_to_leaves(self):
        csg = CallStackGraph(self.df1, self.csi)
        self.assertListEqual(csg.get_paths_to_leaves(0), self.paths_to_leaves_of_0)
        self.assertListEqual(csg.get_paths_to_leaves(5), [self.leaf_nodes_of_5])

    def test_node_depth(self):
        csg = CallStackGraph(self.df1, self.csi)
        nodes = csg.get_nodes()
        df = csg.get_dataframe()

        depth_from_csg = csg.get_depth().to_dict()
        depth_from_nodes = {idx: node.depth for idx, node in nodes.items() if idx >= 0}
        self.assertDictEqual(depth_from_csg, depth_from_nodes)


if __name__ == "__main__":  # pragma: no cover
    unittest.main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest

import pandas as pd

from hta.common.trace import transform_correlation_to_index


class CorrelationTestCase(unittest.TestCase):
    def test_something(self):
        data = [
            [1, 15, -1],
            [2, 16, -1],
            [3, 22, -1],
            [4, -1, -1],
            [5, 15, 7],
            [6, 16, 7],
            [7, 30, 7],
            [8, -1, 7],
        ]
        df = pd.DataFrame(
            data,
            columns=["index", "correlation", "stream"],
            index=[1, 2, 3, 4, 5, 6, 7, 8],
        )
        expected_index_correlation = [5, 6, 0, -1, 1, 2, 0, -1]
        df2 = transform_correlation_to_index(df)
        self.assertListEqual(
            expected_index_correlation, df2["index_correlation"].tolist()
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import random
from enum import Enum
from typing import Dict, List, Optional, Union

import pandas as pd
import plotly.graph_objects as go

from hta.common.trace import Trace
from hta.configs.config import logger
from hta.utils.utils import flatten_column_names, shorten_name


class DeviceType(Enum):
    CPU = 1
    GPU = 2
    ALL = 3


# Define the TraceDir type alias
TraceDir = str


class LabeledTrace:
    """A wrapper class for the Trace class which assigns a label to each trace object.

    Attributes:
        label (str): a label attached to the trace.
        t (Trace): a Trace object that contains the trace data.
        iteration_df (pd.DataFrame): a DataFrame that contains the
    """

    def __init__(
        self,
        label: str = None,
        t: Optional[Trace] = None,
        trace_dir: Optional[str] = None,
    ):
        """Construct a LabeledTrace from either a Trace object or trace files in trace_dir."""
        self.label = label if label else f"t{random.randint(0,10)}"
        if t is not None:
            self.t = t
        elif trace_dir is not None and os.path.isdir(trace_dir):
            self.t = Trace(trace_dir=trace_dir)
        else:
            raise ValueError(
                "either a trace object or a valid trace dir must be provided in LabeledTrace.__init__()"
            )
        self.t.parse_traces()

        self.s_map = pd.Series(self.t.symbol_table.get_sym_id_map())
        self.s_tab = pd.Series(self.t.symbol_table.get_sym_table())
        self.iteration_df = self._extract_iterations()

    def ranks(self) -> List[int]:
        """Get all available ranks."""
        return sorted(self.t.traces.keys())

    def iterations(self) -> List[int]:
        """Get all iterations"""
        return sorted(self.iteration_df["iteration"].values.tolist())

    def _extract_iterations(self) -> pd.DataFrame:
        """Extract the iterations from the symbol map"""
        s_map = pd.Series(self.t.symbol_table.get_sym_id_map())
        iteration_df = s_map[s_map.index.str.startswith("ProfilerStep")].reset_index()
        iteration_df["iteration"] = iteration_df["index"].apply(
            lambda s: int(s.replace("ProfilerStep#", ""))
        )
        iteration_df.rename(columns={0: "id", "index": "symbol"}, inplace=True)
        return iteration_df

    def extract_ops(
        self,
        rank: Optional[Union[int, List[int]]] = None,
        iteration: Optional[Union[int, List[int]]] = None,
        device_type: DeviceType = DeviceType.ALL,
    ) -> pd.DataFrame:
        """
        Description:
            Extract operators/kernels for a specific ranks, iterations, and device type (CPU, GPU, or both)

        Args:
            rank (Optional[Union[int, List[int]]]): Specify which rank(s) to use for the comparison.
                Default: None (use the first rank of the traces.)
            iteration (Optional[Union[int, List[int]]]): Specify which iterations(s) to use for the comparison.
                Default: None (use the first iteration of the traces.)
            device_type (DeviceType): Specify whether to compare CPU operators or GPU kernels or both.
                device_type = DeviceType.CPU    - compare CPU operators only
                device_type = DeviceType.GPU    - compare GPU kernels only
                device_type = DeviceType.ALL    - Compare both CPU operators and GPU kernels.
                Default: DeviceType.ALL

        Returns:
            pd.DataFrame
                A DataFrame of the trace records filtered by ranks, iterations, and device type.
        """
        # Set up and validate the ranks argument
        _ranks = self.ranks()
        if rank is None:
            if len(_ranks) >= 1:
                ranks = _ranks[:1]
            else:
                raise ValueError(f"Trace {self.label} doesn't have any rank.")
        else:
            if isinstance(rank, int):
                ranks = [rank]
            elif isinstance(rank, list):
                ranks = rank
            else:
                raise ValueError(f"Invalid type for argument `rank` {rank}.")

        for _rank in ranks:
            if _rank not in _ranks:
                raise ValueError(f"Invalid argument rank={rank}.")

        # Set up and validate the iteration arguments
        _iterations = self.iterations()
        if iteration is None:
            if len(_iterations) >= 1:
                iterations = _iterations[:1]
            else:
                raise ValueError(f"Trace {self.label} doesn't have any iteration.")
        else:
            if isinstance(iteration, int):
                iterations = [iteration]
            elif isinstance(iteration, list):
                iterations = iteration
            else:
                raise ValueError(f"Invalid type for argument `iteration` {iteration}.")

        for iter_nbr in iterations:
            if iter_nbr not in _iterations:
                raise ValueError(f"Invalid argument iteration={iter_nbr}.")

        # Filter the trace by ranks
        if len(ranks) == 1:
            df_rank = self.t.get_trace(ranks[0])
        else:
            df_rank = pd.concat(
                [self.t.get_trace(r) for r in ranks],
                axis=0,
                keys=_ranks,
                names=["rank", "idx"],
            ).reset_index()

        # Filter the trace by iterations
        df_iter = df_rank[df_rank["iteration"].isin(iterations)]

        # Filter the trace by device type
        if device_type == DeviceType.CPU:
            df = df_iter[df_iter["stream"].eq(-1)]
        elif device_type == DeviceType.GPU:
            df = df_iter[df_iter["stream"].ne(-1)]
        else:
            df = df_iter

        return df

    def get_ops_summary(self, ops: pd.DataFrame) -> pd.DataFrame:
        """
        Description:
            Get the summary of a set of ops.

        Args:
            ops (pd.DataFrame): a set of operations/kernels

        Returns:
            A summary of the counts and total duration of ops grouped by categories and names.

                    cat 	        name 	            short_name 	    counts 	total_duration 	cat_id 	name_id
            0 	user_annotation 	nccl:all_reduce 	nccl:all_reduce 	1 	67 	            10 	    3
            1 	user_annotation 	ProfilerStep#1010 	ProfilerStep#1010 	1 	122149 	        10 	    17
            2 	cpu_op 	            aten::as_strided 	aten::as_strided 	1 	0 	            12 	    0
        """
        s_tab = self.s_tab
        df = (
            ops[["cat", "name", "dur"]]
            .groupby(["cat", "name"])
            .aggregate(["count", "sum"])
        )
        df.columns = ["_".join(col).rstrip("_") for col in df.columns.values]
        df.reset_index(inplace=True)
        df.rename(
            columns={
                "name": "name_id",
                "cat": "cat_id",
                "dur_count": "counts",
                "dur_sum": "total_duration",
            },
            inplace=True,
        )
        df["cat"] = df["cat_id"].apply(lambda i: s_tab[i])
        df["name"] = df["name_id"].apply(lambda i: s_tab[i])
        df["short_name"] = df["name"].apply(lambda n: shorten_name(n))
        df = df[
            [
                "cat",
                "name",
                "short_name",
                "counts",
                "total_duration",
                "cat_id",
                "name_id",
            ]
        ]

        return df


def _trace_argument_adapter(
    t: Union[LabeledTrace, Trace, TraceDir], default_label: str
) -> LabeledTrace:
    """A helper function to construct a LabeledTrace from several argument types."""
    lt: LabeledTrace
    if isinstance(t, TraceDir):
        lt = LabeledTrace(label=default_label, trace_dir=t)
    elif isinstance(t, Trace):
        lt = LabeledTrace(label=default_label, t=t)
    elif isinstance(t, LabeledTrace):
        lt = t
    else:
        raise ValueError(f"Invalid argument type for ({t}).")
    return lt


class TraceDiff:
    @classmethod
    def compare_traces(
        cls,
        control: Union[LabeledTrace, TraceDir],
        test: Union[LabeledTrace, TraceDir],
        control_rank: Optional[Union[int, List[int]]] = None,
        test_rank: Optional[Union[int, List[int]]] = None,
        control_iteration: Optional[Union[int, List[int]]] = None,
        test_iteration: Optional[Union[int, List[int]]] = None,
        device_type: DeviceType = DeviceType.ALL,
        use_short_name: bool = False,
    ) -> pd.DataFrame:
        r"""
        Compare the operators/kernels counts and total duration of two traces.

        Args:
            control (Union[LabeledTrace, Trace, TraceDir]): the control trace.
                A string or Trace object that defines the control trace. Possible values can be:
                    1. a str (TraceDir) that points to parent path of the trace files.
                    2. a Trace object that contains the trace records and metadata.
                    3. a LabeledTrace object, which is a wrapper of the Trace object with a label to identify the trace.

            test (Union[LabeledTrace, Trace, TraceDir]): the test trace.
                Similar to the control trace except it defines the test trace.

            control_rank (Optional[Union[int, List[int]]]): Specify which ranks of the control trace to use.
                This argument can be either a single rank (an integer) or multiple ranks (a list of integers).
                Default: use the first rank of the control traces for comparison

            test_rank (Optional[Union[int, List[int]]]):: Specify which rank of the test trace to use. See ``control_rank`` for usage.
                Default: use the first rank of the test traces for comparison

            control_iteration (Optional[Union[int, List[int]]]): Specify which iteration(s) of the control trace to use for the comparison.
                Default: use the first iteration of the control trace.

            test_iteration (Optional[Union[int, List[int]]]): Specify which iteration(s) of the test trace to use for the comparison.
                Default: use the first iteration of the test trace.

            device_type (DeviceType): Specify whether to compare CPU operators or GPU kernels or both.
                + device_type = DeviceType.CPU compares CPU operators only.
                + device_type = DeviceType.GPU compares GPU kernels only.
                + device_type = DeviceType.ALL compares both CPU operators and GPU kernels.
                Default: DeviceType.ALL

            use_short_name (bool): should the comparison use a shorter name?
                The CPU operator and CUDA kernel name in the trace can be too long to comprehend. The use_short_name argument
                removes the functional arguments, template arguments, and return values so that the name is easy to follow.
                Default: False.

        Returns:
            pd.DataFrame
                A DataFrame that summarizes the difference between the two traces with the following columns:

                + name (or short_name): the operator or kernel name. If ``use_short_name = True``, then short name will be used.
                + control_counts: the number of times an op occurs in the control trace.
                + test_counts: the number of times an op occurs in the test trace.
                + control_total_duration: the total duration of the ops in the control trace.
                + test_total_duration: the total duration of the ops in the test trace.
                + diff_counts: the difference in the counts, i.e. test_counts - control_counts.
                + diff_duration: the difference in total duration, i.e. test_total_duration - control_total_duration.
                + counts_change_categories: the type of changes. '=' (no change); '+' (more ops in test_trace); '-' (fewer ops in test_trace).

        Note:
            The unit for all the duration columns is microsecond (us).
        """
        # Setup control_trace and test_trace
        control_trace = _trace_argument_adapter(control, "Control")
        test_trace = _trace_argument_adapter(test, "Test")

        logger.info(f"comparing traces: {control_trace.label} and {test_trace.label}")
        if control_trace.label == test_trace.label:
            test_trace.label = f"{test_trace.label}_control"
            test_trace.label = f"{test_trace.label}_test"
            logger.warn(
                f"The two traces have the same label. change test_trace's label to {test_trace.label}"
            )
        control_label = control_trace.label
        test_label = test_trace.label

        # Determine which column use to group the operators
        col_name = "short_name" if use_short_name else "name"
        control_trace_summary = (
            control_trace.get_ops_summary(
                control_trace.extract_ops(control_rank, control_iteration, device_type)
            )
            .groupby(col_name)[["counts", "total_duration"]]
            .sum()
        )

        test_trace_summary = (
            test_trace.get_ops_summary(
                test_trace.extract_ops(test_rank, test_iteration, device_type)
            )
            .groupby(col_name)[["counts", "total_duration"]]
            .sum()
        )

        comp = pd.concat(
            [control_trace_summary, test_trace_summary],
            axis=1,
            join="outer",
            keys=[control_label, test_label],
        )
        comp.fillna(0, inplace=True)
        flatten_column_names(comp)

        comp["diff_counts"] = (
            comp[f"{test_label}_counts"] - comp[f"{control_label}_counts"]
        )
        comp["diff_duration"] = (
            comp[f"{test_label}_total_duration"]
            - comp[f"{control_label}_total_duration"]
        )
        comp["counts_change_categories"] = comp["diff_counts"].apply(
            lambda c: "+" if c > 0 else "-" if c < 0 else "="
        )

        return comp

    @classmethod
    def ops_diff(
        cls,
        control: Union[LabeledTrace, TraceDir],
        test: Union[LabeledTrace, TraceDir],
        control_rank: Optional[Union[int, List[int]]] = None,
        test_rank: Optional[Union[int, List[int]]] = None,
        control_iteration: Optional[Union[int, List[int]]] = None,
        test_iteration: Optional[Union[int, List[int]]] = None,
        device_type: DeviceType = DeviceType.ALL,
    ) -> Dict[str, List[str]]:
        r"""
        Get the operator difference between two traces.

        Args:
            control (Union[LabeledTrace, Trace, TraceDir]): The control trace.
                A string or Trace object that defines the control trace. A possible value can be:
                    1. a str (TraceDir) that points to parent path of the trace files.
                    2. a Trace object that contains the trace records and metadata.
                    3. a LabeledTrace object, which is a wrapper of the Trace object with a label to identify the trace.

            test (Union[LabeledTrace, Trace, TraceDir]): The test trace.
                Similar to the control trace except it defines the test trace.

            control_rank (Optional[Union[int, List[int]]]): Specify which ranks
                of the control trace to use. This argument can be either a single
                rank (an integer) or multiple ranks (a list of integers).
                Default: use the first rank of the control traces for comparison

            test_rank (Optional[Union[int, List[int]]]):: Specify which rank
                of the test trace to use. See control_rank for usage.
                Default: use the first rank of the test traces for comparison

            control_iteration (Optional[Union[int, List[int]]]): Specify which
                iteration(s) of the control trace to use for the comparison.
                Default: use the first iteration of the control trace.

            test_iteration (Optional[Union[int, List[int]]]): Specify which
                iteration(s) of the test trace to use for the comparison.
                Default: use the first iteration of the test trace.

            device_type (DeviceType): Specify whether to compare CPU operators
                or GPU kernels or both. DeviceType.CPU, DeviceType.GPU, DeviceType.ALL
                compares CPU operators only, GPU kernels only, both CPU operators
                and GPU kernels respectively. Default: DeviceType.ALL

        Returns:
            Dict[str, List[str]]

                The operator/kernel changes are divided into five types:
                    + added: ops which are absent in control_trace but exist in test_trace.
                    + deleted: ops which exist in control_trace but are absent in test_trace.
                    + increased: ops which exist in both traces but occur more times in test_trace.
                    + decreased: ops which exist in both traces but occur fewer times in test_trace.
                    + unchanged: ops which exist in both traces and occur the same number of times.
        """
        control_trace = _trace_argument_adapter(control, "Control")
        test_trace = _trace_argument_adapter(test, "Test")
        df = cls.compare_traces(
            control_trace,
            test_trace,
            control_rank,
            test_rank,
            control_iteration,
            test_iteration,
            device_type,
        )
        col_control = f"{control_trace.label}_counts"
        col_test = f"{test_trace.label}_counts"
        col_diff = "diff_counts"
        return {
            "added": df.loc[df[col_control].eq(0) & df[col_test].gt(0)].index.tolist(),
            "deleted": df.loc[
                df[col_control].gt(0) & df[col_test].eq(0)
            ].index.tolist(),
            "increased": df.loc[
                df[col_control].gt(0) & df[col_diff].gt(0)
            ].index.tolist(),
            "decreased": df.loc[df[col_test].gt(0) & df[col_diff].lt(0)].index.tolist(),
            "unchanged": df.loc[df[col_test].gt(0) & df[col_diff].eq(0)].index.tolist(),
        }

    @classmethod
    def visualize_counts_diff(
        cls,
        df: pd.DataFrame,
        show_image: bool = True,
        export_image_path: Optional[str] = None,
    ) -> None:
        r"""
        Visualize the changes in trace ops count using the output from ``compare_traces``.

        Args:
            df (pd.DataFrame): the result obtained from ``TraceDiff.compare_traces``.
            show_image (bool): set to True to display the image. Default = True.
            export_image_path (str): location where the image is saved.

        Returns:
            None
        """
        labels: List[str] = [
            col.replace("_total_duration", "")
            for col in df.columns
            if col.endswith("_total_duration")
        ]
        assert len(labels) >= 2

        fig = go.Figure(
            data=[
                go.Bar(name=labels[0], x=df.index, y=df[f"{labels[0]}_counts"]),
                go.Bar(name=labels[1], x=df.index, y=df[f"{labels[1]}_counts"]),
                go.Bar(name="Difference", x=df.index, y=df["diff_counts"]),
            ]
        )
        fig.update_layout(barmode="group", title_text="Ops Count Comparison")
        if show_image:
            fig.show()
        if export_image_path:
            fig.write_image(export_image_path)

    @classmethod
    def visualize_duration_diff(
        cls,
        df: pd.DataFrame,
        show_image: bool = True,
        export_image_path: Optional[str] = None,
    ) -> None:
        r"""
        Visualize the changes in trace ops duration using the output from ``compare_traces``.

        Args:
            df (pd.DataFrame): the result obtained from ``TraceDiff.compare_traces``.
            show_image (bool): set to True to display the image. Default = True.
            export_image_path (str): location where the image is saved.

        Returns:
            None
        """
        labels: List[str] = [
            col.replace("_total_duration", "")
            for col in df.columns
            if col.endswith("_total_duration")
        ]
        assert len(labels) >= 2

        fig = go.Figure(
            data=[
                go.Bar(
                    name=labels[0],
                    x=df.index,
                    y=df[f"{labels[0]}_total_duration"],
                ),
                go.Bar(
                    name=labels[1],
                    x=df.index,
                    y=df[f"{labels[1]}_total_duration"],
                ),
                go.Bar(name="Difference", x=df.index, y=df["diff_duration"]),
            ]
        )
        fig.update_layout(
            barmode="group",
            title_text="Ops Duration Comparison",
            yaxis_title="Total Duration (us)",
            autosize=True,
            width=1000,
            height=1000,
        )
        if show_image:  # pragma: no cover
            fig.show()
        if export_image_path:  # pragma: no cover
            fig.write_image(export_image_path)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import functools
import logging
from collections import namedtuple
from enum import Enum
from time import perf_counter
from typing import Dict, List, NamedTuple, Optional

import numpy as np
import pandas as pd

from hta.common.trace import Trace

NON_EXISTENT_NODE_INDEX = -2
NULL_NODE_INDEX = -1
EVENT_START = 1
EVENT_END = -1


class DeviceType(Enum):
    UNKNOWN = 0
    CPU = 1
    GPU = 2


def infer_device_type(df: pd.DataFrame) -> DeviceType:
    """Infer the device type based on trace data.

    Args:
        df (pd.DataFrame): the filtered dataframe for a single thread/stream

    Returns:
        DeviceType: the type of device
    """
    streams = df["stream"].unique()
    device_type: DeviceType = DeviceType.UNKNOWN

    if len(streams) > 0:
        if np.all(np.greater(streams, 0)):
            device_type = DeviceType.GPU
        elif np.all(np.less(streams, 0)):
            device_type = DeviceType.CPU
    return device_type


"""
We break each trace record into two Event objects: the entity starts and the entity ends.
The Event objects are temporary objects for constructing the call stack.
When an entity starts, it is pushed into the call stack;
when an entity ends, it is popped off the call stack.
In an Event object, the index represents the entity, the type indicates whether the entity
starts or ends at the given time. The duration (<dur>) field is added to resolve cases
where two events may happen at the same time. The time, type, dur, and index are used to
sort the events in the same thread (call stack) using the sorting algorithm provided in
compare_events.
"""
Event = namedtuple("Event", ["idx", "time", "dur", "type"])


def compare_events(x: Event, y: Event) -> int:
    """Compare two events

    Args:
        x, y (Event): the two events to compare

    Returns:
        the ordering of two events
        < 0     x should go first than y
        = 0     same events
        > 0     y should go first than x

    Note:
        There are six cases:
        1. different time:
            Event(idx=1, time=0, dur=10, type=1)
            Event(idx=2, time=2, dur=5, type=1)
            Event(idx=2, time=7, dur=5, type=-1)
            Event(idx=1, time=10, dur=10, type=-1)
        2. same start time, different duration:
            Event(idx=1, time=0, dur=10, type=1)
            Event(idx=2, time=0, dur=5, type=1)
            Event(idx=2, time=5, dur=5, type=-1)
            Event(idx=1, time=10, dur=10, type=-1)
        3. Same end time, different duration:
            Event(idx=1, time=0, dur=10, type=1)
            Event(idx=1, time=10, dur=10, type=-1)
            Event(idx=2, time=10, dur=5, type=1)
            Event(idx=2, time=15, dur=5, type=-1)
        4. same time, one start event, one end event
            Event(idx=1, time=0, dur=10, type=1)
            Event(idx=1, time=10, dur=10, type=-1)
            Event(idx=2, time=10, dur=5, type=1)
            Event(idx=2, time=15, dur=5, type=-1)
        5. same time, same event type, same duration, different index
            Event(idx=1, time=0, dur=10, type=1)
            Event(idx=2, time=0, dur=10, type=1)
            Event(idx=2, time=10, dur=10, type=-1)
            Event(idx=1, time=10, dur=10, type=-1)
        6. same time, same event type, same duration, same index
            The ordering doesn't matter.
    """
    result = x.time - y.time
    if result == 0:
        if x.type == y.type:
            if x.type == EVENT_START:
                if x.dur == y.dur:
                    result = -1 if x.idx < y.idx else 1 if x.idx > y.idx else 0
                else:
                    result = 1 if x.dur < y.dur else -1
            else:
                if x.dur == y.dur:
                    result = 1 if x.idx < y.idx else -1 if x.idx < y.idx else 0
                else:
                    result = -1 if x.dur < y.dur else 1
        else:
            result = 1 if x.type == EVENT_START else -1
    return result


class CallStackNode(NamedTuple):
    """A CallStackNode object captures the connections between entities in the traces.

    Each CallStackNode maps to a unique trace entity, which is an abstraction for operators,
    kernels, functions, user annotations, module components, or any other entities in the traces.

    Each entity is represented by its index (i.e., ID) in the trace data representation.
    A CallStackNode does not store all the attributes of a trace entity. To get the entity's
    attributes, we use CallStackNode's index to query the trace data.

    We assume all entity indices are non-negative integers and use the sentinel value
    NULL_NODE_INDEX for a non-existent entity.

    Attributes:
        parent (int) : the index of the parent.
        depth (int) : the depth on the call stack
        children (List[int]) : the indices of the entities called by this entity of this node.
    """

    parent: int = NULL_NODE_INDEX
    depth: int = -1
    children: List[int] = []


class CallStackIdentity(NamedTuple):
    """A CallStackIdentity object keeps the identity data of a CallStackGraph object.

    Attributes:
        self.rank (int) : the trainer rank.
        self.pid (int) : the process ID of the traces used to construct the CallStack.
        self.tid (int) : the thread ID or stream ID of the trace used to construct the CallStack.
        self.device_type (DeviceType) : the type of the device on which the thread/stream is executed.
    """

    rank: int = -1
    pid: int = -1
    tid: int = -1


class CallStackGraph:
    """A CallStackGraph object tracks the call stacks constructed from the execution traces of
    a single CPU thread or GPU stream.

    Attributes:
        identity (CallStackIdentity) : the identity of this CallStackGraph object.
        df (pd.DataFrame) : the dataframe used to generate this CallStackGraph object.
        nodes (Dict[int, node]): a map from a trace entity's index to a CallStackNode object.
        device_type (DeviceType) : what type of device that the call stack resides.
        correlations (pd.Series) : a Series that maps a node index to the index of a correlated node.

    Notes:
    + Because the kernels on a GPU has only one level, we don't construct a call stack for GPU kernels.
    """

    def __init__(self, df: pd.DataFrame, identity: CallStackIdentity) -> None:
        """Construct an empty graph."""
        self.df = df
        self.identity: CallStackIdentity = identity
        self.device_type: DeviceType = infer_device_type(df)
        self.nodes: Dict[int, CallStackNode] = {}
        self.correlations: pd.Series = None
        self.depth: pd.Series = None
        self._construct_call_stack_graph(df)
        self._compute_depth()

    def __repr__(self):
        ret = "\n"
        for key, item in self.nodes.items():
            ret = ret + f"    {key}: {item}\n"

        return f"CallStackGraph({ret})"

    def _construct_call_stack_graph(self, df) -> None:
        """Construct the call stack from the trace.

        In this function, we assume:
        (1) the traces are from a single thread/stream and therefore
        (2) there is no overlap between the time intervals of the entities on the same level of the graph.

        We skip the call graph construction for GPU streams because the kernels on a single stream is just a list.
        """
        if "index_correlation" not in df.columns:
            raise ValueError(
                "The input DataFrame doesn't have column 'index_correlation'"
            )
        self.correlations = df["index_correlation"]

        if self.device_type == DeviceType.GPU:
            return

        self.nodes.clear()
        self.nodes[NULL_NODE_INDEX] = CallStackNode(NULL_NODE_INDEX, -1, [])
        events = []
        df = df[["index", "ts", "dur"]].copy()
        df["end"] = df["ts"] + df["dur"]
        for _, row in df.iterrows():
            events.append(Event(row["index"], row["ts"], row["dur"], EVENT_START))
            events.append(Event(row["index"], row["end"], row["dur"], EVENT_END))
        events.sort(key=functools.cmp_to_key(compare_events))

        stack: List[Event] = []
        for e in events:
            if e.type == EVENT_START:
                if len(stack) > 0:
                    parent_index = stack[-1].idx
                else:
                    parent_index = NULL_NODE_INDEX
                self._add_edge(parent_index, e.idx)
                stack.append(e)
            else:  # e.type == EVENT_END
                if len(stack) > 0:
                    stack.pop(-1)

    def _add_edge(self, parent_index: int, child_index: int) -> None:
        """Add an edge (parent->child) to the graph.
        Args:
            parent_index (int): the index of the parent node.
            child_index (int): the index of the child node.
        """
        if child_index in self.nodes:
            # Based on the single thread assumption, a child node should always come after the parent node.
            logging.error(f"node {child_index} has already existed.")
            return

        if parent_index not in self.nodes:
            # This should only occurs for the root node
            self.nodes[parent_index] = CallStackNode(NULL_NODE_INDEX, 0, [child_index])
        else:
            self.nodes[parent_index].children.append(child_index)

        # The parent node should always exist at this point.
        self.nodes[child_index] = CallStackNode(
            parent_index, self.nodes[parent_index].depth + 1, []
        )

    def get_nodes(self) -> Dict[int, CallStackNode]:
        """Return the nodes of this graph."""
        return self.nodes

    def get_parent(self, idx: int) -> int:
        """Return the parent of a given node <idx>""

        Args:
            idx (int): the index of a node.

        Returns:
            int: the index of the parent node; return -2 if node <idx> is not in the graph.
        """
        if idx in self.nodes:
            return self.nodes[idx].parent
        logging.error(f"node {idx} is not in current CallStackGraph {self.identity}")
        return NON_EXISTENT_NODE_INDEX

    def get_children(self, idx: int) -> List[int]:
        """Return the children of node <idx>"""
        if idx in self.nodes:
            return self.nodes[idx].children
        return []

    def get_path_to_root(self, idx: int) -> List[int]:
        """Get all the node indices along the path from the node <idx> to the root node

        Args:
            idx (int): the index of a given node.

        Returns:
            List[int]: the list of ancestors' indices.
        """
        if idx not in self.nodes:
            return []

        path = [idx]
        while idx >= 0:
            if idx in self.nodes:
                parent = self.nodes[idx].parent
                path.append(parent)
                idx = parent
            else:
                break
        return path

    def get_paths_to_leaves(self, idx: int) -> List[List[int]]:
        """Get all the paths from the node <idx> as the root to leaf nodes.

        Args:
            idx (int): the index of a given node.

        Returns:
            List[List[int]]: the list of paths from node <idx> to leaf nodes.
        """
        paths = []
        curr_path = []

        def _dfs(_idx: int) -> None:
            if _idx not in self.nodes:
                return

            curr_path.append(_idx)
            if not self.nodes[_idx].children:
                paths.append(list(curr_path))
            else:
                for child in self.nodes[_idx].children:
                    _dfs(child)
            curr_path.pop()

        _dfs(idx)
        return paths

    def get_leaf_nodes(self, idx: int) -> List[int]:
        """Get all leaf nodes on the sub graph with node <idx> as the root.

        Args:
            idx (int): the index of a given node.

        Returns:
            List[int]: the list of leaves nodes on the sub graph with node <idx> as the root.
        """
        return [path[-1] for path in self.get_paths_to_leaves(idx)]

    def get_dataframe(self) -> pd.DataFrame:
        """Get the trace dataframe for this stack"""
        return self.df

    def _compute_depth(self) -> None:
        """Add the depth information to the DataFrame"""
        if self.device_type == DeviceType.GPU:
            self.depth = pd.Series(
                data=np.full(self.correlations.size, -1),
                index=self.correlations.index,
                name="depth",
                copy=True,
            )
        else:  # self.device_type == DeviceType.CPU:
            self.depth = pd.Series(
                data={idx: node.depth for idx, node in self.nodes.items() if idx >= 0},
                name="depth",
                copy=True,
            )

    def get_depth(self) -> pd.Series:
        """Get the depth for all valid node

        Return:
            a Series with the node index as index and depth as the data
        """
        return self.depth


class CallGraph:
    """
    A CallGraph represents the entire set of traces with a set of CallStackGraph
    objects.

    The execution of a distributed training job can be abstracted as a hierarchical
    organization of CallStackGraph object, which abstracts the execution of a single
    thread/stream. The hierarchical structure is as follows:
    + distribute training job
      ++ trainer
         +++ process
           ++++ thread/stream
             ++++ a sequence of entity events - represented with a CallStackGraph object

    Because there are possible relationship links between two or more CallStackGraph objects,
    such as Cuda Kernel launches, AllToAll communications, etc., a CallStackGraph object
    includes all CallStackGraph objects in the trace and provides further further query and statistic APIs.

    Attributes:
        trace_data (Trace) : the trace data represented in a Trace object, which
            contains multiple DataFrame objects mapping the traces of each trainer.
        call_stacks (List[CallStackGraph]) : a list of per-thread CallStackGraph objects.
        mapping (pd.DataFrame) : the mapping from CallStackIdentity to CallStackGraph using a DataFrame
    """

    def __init__(self, trace: Trace, ranks: Optional[List[int]] = None) -> None:
        """Construct a CallGraph from a Trace object <trace_data>

        Args:
            trace (Trace): the trace data used to construct this CallGraph object.
            ranks (List[int]) : filter the traces using the given set of ranks. Using all ranks if None.
        Raises:
            ValueError: the trace data is invalid.
        """
        self.trace_data: Trace = trace
        self.mapping: pd.DataFrame = pd.DataFrame()
        self.call_stacks: List[CallStackGraph] = []

        _ranks = [k for k in trace.get_all_traces()] if ranks is None else ranks
        self._construct_call_graph(_ranks)

    def _construct_call_graph(self, ranks: List[int]) -> None:
        """
        Construct the call graph from the traces of a distributed training job.

        Args:
            ranks (List[int]) : a list ranks to select traces for construct the call stacks.
        """
        call_stack_ids: List[CallStackIdentity] = []
        t0 = perf_counter()
        # construct a call stack graph for each thread/stream
        for rank in ranks:
            df = self.trace_data.get_trace(rank)
            for pid, pid_group in df.groupby(by="pid"):
                for tid, tid_group in pid_group.groupby(by="tid"):
                    csi = CallStackIdentity(rank, pid, tid)
                    csg = CallStackGraph(tid_group, csi)
                    self.call_stacks.append(csg)
                    call_stack_ids.append(csi)

        t1 = perf_counter()
        logging.debug(
            f"Completed constructing call stack graph for in {t1 - t0:.3} seconds"
        )

        # build a map from call stack meta data to call stack objects
        self.mapping = pd.DataFrame(
            {
                "rank": [csi.rank for csi in call_stack_ids],
                "pid": [csi.pid for csi in call_stack_ids],
                "tid": [csi.tid for csi in call_stack_ids],
                "csg_index": range(len(self.call_stacks)),
            }
        )

        # add depth information to the data frame
        for rank in ranks:
            call_stack_indices = self.mapping[self.mapping["rank"].eq(rank)][
                "csg_index"
            ]
            depth = pd.concat(
                [self.call_stacks[idx].get_depth() for idx in call_stack_indices]
            )
            df = self.trace_data.get_trace(rank)
            df["depth"] = depth
        self.mapping.set_index(["rank", "pid", "tid"], inplace=True)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict, List, Optional

import pandas as pd

from hta.common.call_stack import CallGraph
from hta.common.trace import Trace
from hta.configs.config import logger

CUDA_SASS_INSTRUCTION_COUNTER_FLOPS: Dict[str, float] = {
    f"smsp__sass_thread_inst_executed_op_{op}_pred_on.sum": (2 if "fma" in op else 1)
    for op in ["ffma", "fmul", "fadd", "hfma", "hmul", "hadd", "dfma", "dmul", "dadd"]
}


class CuptiCounterAnalysis:
    def __init__(self):
        pass

    @classmethod
    def _get_counter_data_with_operators_for_rank(
        cls,
        t: "Trace",
        rank: int,
        cg: CallGraph,
    ) -> Optional[pd.DataFrame]:
        sym_table = t.symbol_table.get_sym_table()
        sym_index = t.symbol_table.get_sym_id_map()

        cuda_profiler_cat = sym_index.get("cuda_profiler_range")
        cuda_runtime_cat = sym_index.get("cuda_runtime")
        cpu_op_cat = sym_index.get("cpu_op")
        # TODO:bcoutinho handle missing symbols?
        kernel_launch_sym = sym_index.get("cudaLaunchKernel")

        # Get trace for a rank
        trace_df: pd.DataFrame = t.get_trace(rank)

        # Get cuda profiler events
        gpu_kernels = trace_df[trace_df["cat"].eq(cuda_profiler_cat)].reset_index(
            drop=True
        )

        # call stacks of interest are all on CPU
        call_stacks_idxs = cg.mapping.query(
            f"rank == {rank} and pid != 0 and tid != 0"
        ).csg_index.values

        # merge with runtime events from call stacks
        leaf_nodes: List[int] = []
        for i in call_stacks_idxs:
            leaf_nodes.extend(cg.call_stacks[i].get_leaf_nodes(-1))
        leaf_df = trace_df.loc[leaf_nodes]

        kernel_launches = leaf_df[leaf_df["cat"].eq(cuda_runtime_cat)]
        kernel_launches = kernel_launches[
            kernel_launches["name"].eq(kernel_launch_sym)
        ].reset_index(drop=True)

        # We map 1:1 each kernel launch with gpu kernel
        # so the number of each should be equal

        if len(kernel_launches) != len(gpu_kernels):
            logger.error(
                "Number of kernels launches and kernels do not match for"
                f" rank {rank}\n"
                f" kernel launches ({len(kernel_launches)})"
                f" kernels ({len(gpu_kernels)})"
            )
            return None

        # Now, we do the merge using index only
        gpu_kernels = gpu_kernels.merge(
            kernel_launches[["index"]],
            left_index=True,
            right_index=True,
            how="inner",
            suffixes=[None, "_runtime"],  # adds a column index_runtime
        )

        # Add the op_stack as an array
        def get_opstack(row: pd.Series) -> List[int]:
            for i in call_stacks_idxs:
                op_stack = cg.call_stacks[i].get_path_to_root(row.index_runtime)[:-2]
                if len(op_stack) > 0:
                    return op_stack
            return []

        gpu_kernels["op_stack"] = gpu_kernels.apply(get_opstack, axis=1)

        def get_top_or_bottom_op(ops: List[int], top: bool) -> int:
            # only get ops that are cpu_op
            filterd_ops = [op for op in ops if trace_df.loc[op]["cat"] == cpu_op_cat]
            if len(filterd_ops) < 1:
                return -1
            return filterd_ops[-1 if top else 0]

        gpu_kernels["top_level_op"] = gpu_kernels["op_stack"].apply(
            lambda ops: get_top_or_bottom_op(ops, top=True)
        )
        gpu_kernels["bottom_level_op"] = gpu_kernels["op_stack"].apply(
            lambda ops: get_top_or_bottom_op(ops, top=False)
        )

        # add back strings for readability
        for col in ["cat", "name"]:
            gpu_kernels[col] = gpu_kernels[col].apply(
                lambda i: sym_table[i] if (i > 0 and i < len(sym_table)) else ""
            )
        for col in ["top_level_op", "bottom_level_op"]:
            gpu_kernels[col] = gpu_kernels[col].apply(
                lambda op: sym_table[trace_df.loc[op]["name"]] if op >= 0 else ""
            )

        def stringify_op_stack(ops: List[int]) -> List[str]:
            return [sym_table[trace_df.loc[op]["name"]] for op in ops]

        gpu_kernels["op_stack"] = gpu_kernels["op_stack"].apply(stringify_op_stack)

        return gpu_kernels

    @classmethod
    def get_counter_data_with_operators(
        cls,
        t: "Trace",
        ranks: Optional[List[int]] = None,
    ) -> List[pd.DataFrame]:
        """Correlates the Kernel counter events with pytorch operators using
        the callgraph.
        Args:
            t (Trace): trace object
            ranks (List[int]): List of ranks on which to run the analysis. Default = [0].
        Returns:
            A list of dataframes, one per rank, containing kernel name,
            op_stack (operator stack), top and bottom level op, and columns
            for individual performance counters.

        For more details see `get_counter_data_with_operators` in `trace_analysis.py`,
        or read more here https://github.com/facebookresearch/HolisticTraceAnalysis/issues/29
        """

        if ranks is None or len(ranks) == 0:
            ranks = [0]

        sym_index = t.symbol_table.get_sym_id_map()
        if "cuda_profiler_range" not in sym_index.keys():
            logger.warning(
                "Could not find events of 'cuda_profiler_range' category "
                "Please check if you ran CUPTI profiler mode correctly"
            )
            return []

        cg = CallGraph(t)

        result_list = [
            cls._get_counter_data_with_operators_for_rank(t=t, rank=rank, cg=cg)
            for rank in ranks
        ]

        return [k for k in result_list if k is not None]



# Copyright (c) Meta Platforms, Inc. and affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict, List, Optional

import pandas as pd

from hta.common.trace import Trace
from hta.configs.config import logger
from hta.utils.utils import get_kernel_type, get_memory_kernel_type, KernelType


class TraceCounters:
    def __init__(self):
        pass

    @classmethod
    def _get_queue_length_time_series_for_rank(
        cls, t: "Trace", rank: int
    ) -> Optional[pd.DataFrame]:
        """
        Returns an (optional) dataframe with time series for the queue length
        on a CUDA streams within requested rank.

        Queue length is defined as the number of outstanding CUDA operations on a stream
        The value of the queue length is:
        1. Incremented when a CUDA runtime operation enqueues a kernel on a stream.
        2. Decremented when a CUDA kernel/memcopy operation executes on a stream.

        Args:
            t (Trace): Input trace data structure.
            rank (int): rank to generate the time series for.

        Returns:
            Optional[pd.DataFrame]
                Returns an (optional) dataframe containing time series with the following
                columns: ts (timestamp), pid, tid (of corresponding GPU, stream), stream and
                queue_length.

                Note that each row or timestamp denotes a change in the value of the
                time series. The value remains constant until the next timestamp.
                In essence, it can be thought of as a step function.
        """
        # get trace for a rank
        trace_df: pd.DataFrame = t.get_trace(rank)

        # cudaLaunchKernel, cudaMemcpyAsync, cudaMemsetAsync
        sym_index = t.symbol_table.get_sym_id_map()
        cudaLaunchKernel_id = sym_index.get("cudaLaunchKernel", None)
        cudaMemcpyAsync_id = sym_index.get("cudaMemcpyAsync", None)
        cudaMemsetAsync_id = sym_index.get("cudaMemsetAsync", None)

        # CUDA Runtime events that may launch kernels
        # - filter events that have a correlated kernel event only.
        runtime_calls: pd.DataFrame = trace_df.query(
            "((name == @cudaMemsetAsync_id) or (name == @cudaMemcpyAsync_id) or "
            " (name == @cudaLaunchKernel_id))"
            "and (index_correlation > 0)"
        ).copy()
        runtime_calls.drop(["stream", "pid", "tid"], axis=1, inplace=True)
        runtime_calls["queue"] = 1

        # GPU kernel events
        gpu_kernels = trace_df[trace_df["stream"].ne(-1)].copy()
        gpu_kernels["queue"] = -1

        # Concat the series of runtime launch events and GPU kernel events
        merged_df = (
            pd.concat(
                [
                    # use the pid, tid and cuda stream from the correlated GPU event.
                    runtime_calls.join(
                        gpu_kernels[["stream", "pid", "tid", "correlation"]].set_index(
                            "correlation"
                        ),
                        on="correlation",
                    ),
                    gpu_kernels,
                ]
            )
            .sort_values(by="ts")
            .set_index("index")
        )

        result_df_list = []
        for stream, stream_df in merged_df.groupby("stream"):
            logger.debug(f"Processing queue_length for rank {rank}, stream {stream}")
            stream_df["queue_length"] = stream_df["queue"].cumsum()
            result_df_list.append(stream_df)

        return (
            pd.concat(result_df_list)[["ts", "pid", "tid", "stream", "queue_length"]]
            if len(result_df_list) > 0
            else None
        )

    @classmethod
    def get_queue_length_time_series(
        cls,
        t: "Trace",
        ranks: Optional[List[int]] = None,
    ) -> Dict[int, pd.DataFrame]:
        """
        Returns a dictionary of rank -> time series for the queue length of a CUDA stream.

        Queue length is defined as the number of outstanding CUDA operations on a stream
        The value of the queue length is:
        1. Incremented when a CUDA runtime operation enqueues a kernel on a stream.
        2. Decremented when a CUDA kernel/memcopy operation executes on a stream.

        Args:
            t (Trace): Input trace data structure.
            rank (int): rank to perform this analysis for.

        Returns:
            Dict[int, pd.DataFrame]:
                A dictionary of rank -> time series with the queue length of each CUDA stream.
                Each dataframe contains a time series consisting of the following columns:
                ts (timestamp), pid, tid (of corresponding GPU, stream), stream and queue_length.

                Note that each row or timestamp shows a change in the value of the
                time series. The value remains constant until the next timestamp.
                In essence, it can be thought of as a step function.
        """
        if ranks is None or len(ranks) == 0:
            ranks = [0]

        logger.info(
            "Please note that the time series only contains points "
            "when the value changes. Once a values is observed the time series "
            "stays constant until the next update."
        )

        result = {
            rank: TraceCounters._get_queue_length_time_series_for_rank(t, rank)
            for rank in ranks
        }
        return dict(filter(lambda x: x[1] is not None, result.items()))

    @classmethod
    def get_queue_length_summary(
        cls,
        t: "Trace",
        ranks: Optional[List[int]] = None,
    ) -> Optional[pd.DataFrame]:
        """
        Returns an (optional) dataframe with queue length statistics per CUDA stream and rank.

        Args:
            t (Trace): Input trace data structure.
            ranks (list of int): ranks to perform this analysis.

        Returns:
            Optional[pd.DataFrame]
                An (optional) dataframe containing the summary statistics of queue length per
                stream and rank.
        """
        if ranks is None or len(ranks) == 0:
            ranks = [0]

        results_list: List[pd.DataFrame] = []

        for rank, rank_df in TraceCounters.get_queue_length_time_series(
            t, ranks
        ).items():
            rank_df["rank"] = rank
            result = (
                rank_df[["rank", "stream", "queue_length"]]
                .groupby(["rank", "stream"])
                .describe()
            )
            results_list.append(result)
        return pd.concat(results_list) if len(results_list) > 0 else None

    @classmethod
    def _get_memory_bw_time_series_for_rank(
        cls, t: "Trace", rank: int
    ) -> Optional[pd.DataFrame]:
        """
        Returns time series for the memory bandwidth of memory copy and memory set operations
        for specified rank.

        Args:
            t (Trace): Input trace data structure.
            rank (int): rank to generate the time series for.

        Returns:
            Optional[pd.DataFrame]
                Returns an (optional) dataframe with time series for the memory bandwidth.
                The dataframe returned contains time series with columns:
                ts (timestamp), pid (of corresponding GPU), name of memory copy type
                and memory_bw_gbps (memory bandwidth in GB/sec).
        """
        # get trace for a rank
        trace_df: pd.DataFrame = t.get_trace(rank)
        sym_table = t.symbol_table.get_sym_table()

        gpu_kernels = trace_df[trace_df["stream"].ne(-1)].copy()
        gpu_kernels["kernel_type"] = gpu_kernels[["name"]].apply(
            lambda x: get_kernel_type(sym_table[x["name"]]), axis=1
        )

        memcpy_kernels = gpu_kernels[
            gpu_kernels.kernel_type == KernelType.MEMORY.name
        ].copy()
        memcpy_kernels["name"] = memcpy_kernels[["name"]].apply(
            lambda x: get_memory_kernel_type(sym_table[x["name"]]), axis=1
        )

        # In case of 0 us duration events round it up to 1 us to avoid -ve values
        # see https://github.com/facebookresearch/HolisticTraceAnalysis/issues/20
        memcpy_kernels.loc[memcpy_kernels.dur == 0, ["dur"]] = 1

        membw_time_series_a = memcpy_kernels[["ts", "name", "pid", "memory_bw_gbps"]]
        membw_time_series_b = memcpy_kernels[
            ["ts", "name", "dur", "pid", "memory_bw_gbps"]
        ].copy()

        # The end events have timestamps = start timestamp + duration
        membw_time_series_b.ts = membw_time_series_b.ts + membw_time_series_b.dur
        membw_time_series_b.memory_bw_gbps = -membw_time_series_b.memory_bw_gbps

        membw_time_series = pd.concat(
            [
                membw_time_series_a,
                membw_time_series_b[["ts", "pid", "name", "memory_bw_gbps"]],
            ],
            ignore_index=True,
        ).sort_values(by="ts")

        result_df_list = []
        for _, membw_df in membw_time_series.groupby("name"):
            membw_df.memory_bw_gbps = membw_df.memory_bw_gbps.cumsum()
            result_df_list.append(membw_df)

        if len(result_df_list) == 0:
            return None

        result_df = pd.concat(result_df_list)[["ts", "pid", "name", "memory_bw_gbps"]]
        return result_df

    @classmethod
    def get_memory_bw_time_series(
        cls,
        t: "Trace",
        ranks: Optional[List[int]] = None,
    ) -> Dict[int, pd.DataFrame]:
        """
        Returns a dictionary of rank -> time series for the memory bandwidth.

        Args:
            t (Trace): Input trace data structure.
            ranks (list of int): ranks to perform this analysis for.

        Returns:
            Dict[int, pd.DataFrame]
                Returns a dictionary of rank -> time series for the memory bandwidth.
                The dataframe returned contains time series along with the following columns:
                ts (timestamp), pid (of corresponding GPU), name of memory copy type
                and memory_bw_gbps (memory bandwidth in GB/sec).
        """
        if ranks is None or len(ranks) == 0:
            ranks = [0]

        logger.info(
            "Please note that the time series only contains points "
            "when the value changes. Once a values is observed the time series "
            "stays constant until the next update."
        )
        result = {
            rank: TraceCounters._get_memory_bw_time_series_for_rank(t, rank)
            for rank in ranks
        }
        return dict(filter(lambda x: x[1] is not None, result.items()))

    @classmethod
    def get_memory_bw_summary(
        cls,
        t: "Trace",
        ranks: Optional[List[int]] = None,
    ) -> Optional[pd.DataFrame]:
        """
        Returns an (optional) dataframe containing the summary statistics of memory ops. The
        tracked memory ops are MemcpyDtoH, MemcpyHtoD, MemcpyDtoD and MemSet.

        Args:
            t (Trace): Input trace data structure.
            ranks (list of int): ranks to perform this analysis for.

        Returns:
            Optional[pd.DataFrame]
                An (optional) dataframe containing the summary statistics of the following memory ops:
                MemcpyDtoH, MemcpyHtoD, MemcpyDtoD, MemSet.
        """
        if ranks is None or len(ranks) == 0:
            ranks = [0]

        results_list: List[pd.DataFrame] = []

        for rank, rank_df in TraceCounters.get_memory_bw_time_series(t, ranks).items():
            rank_df["rank"] = rank
            # Exclude the 0 points in time series
            rank_df = rank_df[rank_df.memory_bw_gbps > 0]

            result = (
                rank_df[["rank", "name", "memory_bw_gbps"]]
                .groupby(["rank", "name"])
                .describe()
            )
            results_list.append(result)
        return pd.concat(results_list) if len(results_list) > 0 else None
